/* Copyright (c) 2010-2020 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* Convert various bitmap formats to EPS, PDF or PNG, with
 * emphasis on good compression
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>  /* isatty */

#include "bmp2eps.h"

#ifndef PAPER
#define PAPER 4          /* Default paper size. 0-6 -> A0-A6     */
#endif                   /*                     -1 -> US letter  */

void help();
void versions();
int read_image(struct bitmap *image, FILE *infile,
               int user_trans, unsigned int ctrans[3]);
void write_pdf(FILE *out, struct bitmap *image, int bb[4], int pt);
void write_ps(FILE *out, struct bitmap *image, int bb[4], int pt);
void write_png(FILE *out, struct bitmap *image);
void write_rtl(FILE *out, struct bitmap *image);
void write_svg(FILE *out, struct bitmap *image, int type);
void free_image(struct bitmap *image);

void error_exit(char *msg){
  if (msg) fprintf(stderr,"%s\n",msg);
  exit(1);
}


int main(int argc, char **argv)
{
  int tmp,opt=1;
  int smooth=0,land=0,pt=PAPER,ps=0,ex=0,compress,hex=0,keep24=0;
  int pdf=0,png=0,rtl=0,svg=0,multi_image=0;
  int debug=1,binary=0,distill=0,overwrite=0,rep_shrink=0;
  int user_trans=0,old_flags,ofile;
  int bb[4],dpi_x,dpi_y,c;

  unsigned int ctrans[3],itmp[3];
  unsigned char bbuff[4];
  unsigned int len1,len2;

  char *optp,*file1=NULL,*file2=NULL;
  FILE *infile,*outfile;
  struct bitmap image;
  struct stat stat_buff;


  image.trans=NULL;
  image.dpi_x=image.dpi_y=0;
  image.flags=PREDICT;
  image.bkgd=NULL;
  image.palette=NULL;
  image.apalette=NULL;
  image.pal_len=0;
  image.apal_len=0;
  image.components=0;
  image.smask=NULL;
  image.compress_level=0;
  image.reduce_depth=0;
  bb[0]=bb[1]=bb[2]=bb[3]=0;
  compress=-1;
  ctrans[0]=ctrans[1]=ctrans[2]=255;
  len1=len2=0;
  dpi_x=dpi_y=0;
  ofile=0;
  outfile=NULL;
  
#ifdef USE_ZOPFLI
  image.flags|=ZOPFLI;
#endif
 
  if (strlen(argv[0])>=4){
    if(!strcmp("2pdf",&argv[0][strlen(argv[0])-4])){
      pdf=1;
      binary=1;
    }

    if(!strcmp("2png",&argv[0][strlen(argv[0])-4])){
      png=1;
      binary=1;
      compress=FLATE;
    }
  }

  opt=1;
  optp=argv[opt];

  while (opt<argc){
    switch(*optp){
    case 0:
      opt++;
      optp=argv[opt];
      break;
    case '-':
      if (!*(optp+1)){ /* a bare - */
	opt++;
	if (!file1) {file1="-"; break;}
	else if (!file2) {file2="-"; break;}
	error_exit("Unexpected - in options");
	break;
      }
      if (!strcmp(optp,"--help")) help();
      if (!strcmp(optp,"--version")){
	printf("bmp2eps / bmp2pdf / bmp2png " BMP2EPS_VER
	       " (c) 2002-2020 MJR\n");
	exit(0);
      }
      while(*(++optp)){
        switch(*optp){
        case 'v':
          debug++;
          break;
        case 'h':
	  if ((*(optp+1)=='t')&&(*(optp+2)=='m')&&(*(optp+3)=='l')){
            svg=2;
            optp+=3;
          }
          else help();
          break;
        case 'a':
	  optp++;
          ps=1;
	  switch(*optp){
	  case '0': case '1': case '2': case '3':
	  case '4': case '5':case '6':
	    pt=*optp-'0';
	    break;
	  default:
	    fprintf(stderr,"Invalid option a%c.\n%s -h for usage.\n",
		    *optp,argv[0]);
	    exit(1);
	  }
          break;
        case 'b':
          if (*(optp+1)=='='){
            optp+=2;
            tmp=sscanf(optp,"%2x%2x%2x",itmp,itmp+1,itmp+2);
            if (tmp!=3) tmp=sscanf(optp,"%u,%u,%u",itmp,itmp+1,itmp+2);
	    if ((tmp==0)||(tmp==EOF))
	      image.flags|=NO_BKGD;
            else if ((tmp!=1)&&(tmp!=3)){
              fprintf(stderr,"Invalid use of -b=\n%s -h for usage\n",
		      argv[0]);
              exit(1);
            }
	    bbuff[0]=tmp;
	    bbuff[1]=itmp[0];
	    bbuff[2]=itmp[1];
	    bbuff[3]=itmp[2];
	    image.bkgd=bbuff;
            while(*(optp+1)) optp++; /* Consume the rest of this argument */
          }
          else {
	    fprintf(stderr,"Invalid use of -b=\n%s -h for usage\n",
                    argv[0]);
	    exit(1);
          }
          break;
        case 'B':
          if (*(optp+1)=='='){
            optp+=2;
            tmp=sscanf(optp,"%d,%d,%d,%d",bb,bb+1,bb+2,bb+3);
            if ((tmp!=1)&&(tmp!=2)&&(tmp!=4)){
              fprintf(stderr,"Invalid use of -B=\n%s -h for usage\n",
		      argv[0]);
              exit(1);
            }
            if ((tmp==1)||(tmp==2))  {bb[2]=bb[0];bb[0]=0;}
            if (tmp==2)  {bb[3]=bb[1];bb[1]=0;}
            while(*(optp+1)) optp++; /* Consume the rest of this argument */
	  }else{
	    fprintf(stderr,"Invalid use of -B\n%s -h for usage\n",
                    argv[0]);
	    exit(1);
	  }
	  break;
 	case 'd':
	  distill=1;
	  break;
        case 'D':
          if (*(optp+1)=='='){
            optp+=2;
            tmp=sscanf(optp,"%d",&dpi_x);
            if (tmp!=1){
              fprintf(stderr,"Invalid use of -D=\n%s -h for usage\n",
		      argv[0]);
              exit(1);
            }
            dpi_y=dpi_x;
            while(*(optp+1)) optp++; /* Consume the rest of this argument */
          }
          else dpi_x=dpi_y=-1;
          break;
	case 'e':
	  if ((*(optp+1)=='p')&&(*(optp+2)=='s')){
	    png=0;
	    pdf=0;
	    binary=0;
	    optp+=2;
	  } 
	  else{
	    ex=1;
	    ps=1;
	  }
	  break;
	case 'f':
	  image.compress_level--;
	  break;
        case 'G':
          if (*(optp+1)=='3') compress=CCITT;
          else if (*(optp+1)=='4') compress=CCITTG4;
          else {
	    fprintf(stderr,"Invalid use of -G\n%s -h for usage\n",
                    argv[0]);
	    exit(1);
	  }
          optp++;
          break;
        case 'i':
          image.flags|=PNGI;
          break;
        case 'l':
          land=1;
          break;
	case 'n':
	  compress=0;
	  break;
	case 'N':
          if (*(optp+1)!='='){
	    fprintf(stderr,"Invalid use of -N\n%s -h for usage\n",
                    argv[0]);
	    exit(1);
	  }
          optp+=2;
	  if(sscanf(optp,"%d",&tmp)!=1){
	    fprintf(stderr,"Invalid use of -N\n%s -h for usage\n",
                    argv[0]);
	    exit(1);
	  }
	  image.reduce_depth=tmp;
          while(*(optp+1)) optp++; /* Consume the rest of this argument */
	  break;
	case 'O':
	  overwrite=1;
	  png=1;
	  pdf=0;
	  binary++;
	  if (compress!=FLATEF) compress=FLATE;
	  break;
	case 'o':
	  opt++;
	  file2=argv[opt];
	  ofile=1;
	  optp=file2;
	  while (*(optp+1)) optp++;
	  break;
	case 'P':
          if (*(optp+1)!='='){
	    fprintf(stderr,"Invalid use of -P\n%s -h for usage\n",
                    argv[0]);
	    exit(1);
	  }
          optp+=2;
          if(sscanf(optp,"%d",&tmp)!=1){
	    fprintf(stderr,"Invalid use of -P\n%s -h for usage\n",
                    argv[0]);
	    exit(1);
	  }
          if((tmp<0)||(tmp>15)){
	    fprintf(stderr,"Invalid use of -P\n%s -h for usage\n",
                    argv[0]);
	    exit(1);
	  }
          image.flags&=(~PREDICT);
          image.flags|=(tmp<<P_SHIFT);
          while(*(optp+1)) optp++; /* Consume the rest of this argument */
	  break;
        case 'p':
	  if ((*(optp+1)=='d')&&(*(optp+2)=='f')){
	    pdf=1;
	    png=0;
	    binary++;
	    optp+=2;
	  }
	  else if ((*(optp+1)=='n')&&(*(optp+2)=='g')){
	    png=1;
	    pdf=0;
	    binary++;
	    if (compress!=FLATEF) compress=FLATE;
	    optp+=2;
	  } 
	  else ps=1;
	  break;
	case 'q':
	  debug=0;
	  break;
	case 'r':
	  if ((*(optp+1)=='t')&&(*(optp+2)=='l')){
	    rtl=1;
	    pdf=0;
	    png=0;
	    binary++;
	    optp+=2;
            keep24=1;
	  }
	  else compress=RLE;
	  break;
        case 'R':
          image.flags|=PNG_RAW;
          image.flags|=PNGI;
          break;
        case 's':
          if ((*(optp+1)=='v')&&(*(optp+2)=='g')){
            svg=1;
            optp+=2;
          }
          else smooth=1;
          break;
        case 't':
          if (*(optp+1)=='='){
            optp+=2;
            tmp=sscanf(optp,"%2x%2x%2x",ctrans,ctrans+1,ctrans+2);
            if (tmp!=3) tmp=sscanf(optp,"%u,%u,%u",ctrans,ctrans+1,ctrans+2);
            if ((tmp!=EOF)&&(tmp!=0)&&(tmp!=3)){
              fprintf(stderr,"Invalid use of -t=\n%s -h for usage\n",
		      argv[0]);
              exit(1);
            }
            user_trans=1;
            while(*(optp+1)) optp++; /* Consume the rest of this argument */
          }
          else {
            user_trans=-1;
            image.flags|=NO_TRANS;
          }
          break;
        case 'u':
          pt=-1;
          break;
	case 'V':
	  versions();
	  break;
        case 'x':
          hex=1;
          break;
	case 'X':
	  hex=2;
	  break;
#ifdef USE_ZLIB
	case 'z':
	  compress=FLATE;
	  image.flags&=(~ZOPFLI);
	  break;
#endif
#ifdef USE_LZW
        case 'Z':
          compress=LZW;
          break;
#endif
	case '2':
	  if (*(optp+1)=='4'){
            keep24=1;
            optp++;
            break;
          }
	  else{
	    image.flags|=PSL2;
	    break;
	  }
	case '7':
	  binary--;
	  break;
	case '8':
	  binary++;
	  break;
	case '9':
	  image.compress_level++;
	  break;
        default:
          fprintf(stderr,"Invalid option %c.\n%s -h for usage.\n",
		  *optp,argv[0]);
	  exit(1);
	}
      }
      opt++;
      optp=argv[opt];
      break;
    default:
      if (!file1) file1=argv[opt];
      opt++;
      optp=argv[opt];
      break;
    }
    if (file1) break;
  }

  if (debug>1){
    fprintf(stderr,"bmp2eps " BMP2EPS_VER
#ifdef USE_LZW
	    " with LZW"
#endif
#ifdef USE_ZLIB
	    " with ZLIB"
#endif
#ifdef USE_LIBPNG
	    " with libpng"
#endif
#ifdef USE_ZOPFLI
	    " with zopfli"
#endif
	    "\n");
  }

  if ((overwrite)&&(ofile))
    error_exit("Cannot specify both -o and -O\n");
  
  if ((png)&&(overwrite)&&(file1)&&(strlen(file1)>4)&&
      (!strcasecmp(file1+strlen(file1)-4,".png"))){
    file2=file1;
    rep_shrink=1;
  }
  
  if ((!file2)&&(argc>opt))
    file2=argv[argc-1];
  
  if (file2){
    if (!(strcmp(file2,"-")))
      outfile=stdout;
    else{
      if (file1!=file2){ /* Fail early if cannot open output file */
	outfile=fopen(file2,"wb"); /* need to avoid truncating file2 */
	if(!outfile){
	  fprintf(stderr,"Error, unable to open %s for writing.\n",file2);
	  exit(1);
	}
      }
    }
  }
  else
    outfile=stdout;

  if (binary<0) binary=0;
  if ((binary==1)&&(!file2)&&(isatty(fileno(stdout)))){
    fprintf(stderr,"Refusing to output binary data to a terminal:"
	    " aborting.\n");
    if (pdf)
      fprintf(stderr,"Add a \"-8\" flag to force,"
	      " or \"-7\" for ASCII output\n\n");
    else
      fprintf(stderr,"Add another \"-8\" flag to force.\n\n");
    exit(1);
  }
  if (binary>0) binary=1;

  image.flags|=debug;

  if (compress==-1){
    if (rtl==1) compress=RLE24;
    else compress=DEFAULT_COMPRESS;
  }
  if (ex) image.flags|=EXPAND;
  if (smooth) image.flags|=SMOOTH;
  if (hex==1) image.flags|=HEX;
  if (hex==2) image.flags|=HEX2;
  if (binary) image.flags|=BINARY;
  if (distill) image.flags|=DISTILL;
  if (ps) image.flags|=PAGE;
  if (pdf) image.flags|=PDF;
  if ((png)||(svg)) image.flags|=PNG;
  if (rtl) image.flags|=RTL;
  if (keep24) image.flags|=KEEP24;
  old_flags=image.flags;
  multi_image=0;
  if (pdf) multi_image=1;
  if (ps) multi_image=1;
  if ((png)&&(overwrite)) multi_image=1;
  if ((multi_image==0)&&(opt+1<argc+ofile)){
    fprintf(stderr,"Unexpected arguments present: %s\n",argv[opt+1]);
    exit(1);
  }

  
  do{

    if (file1){
      if (!(strcmp(file1,"-")))
	infile=stdin;
      else{
	infile=fopen(file1,"rb");
	if(!infile){
	  fprintf(stderr,"Error, unable to open %s for reading.\n",file1);
	  exit(1);
	}
	fprintf(stderr,"Processing %s\n",file1);
	if (rep_shrink){
	  if (!stat(file1,&stat_buff))
	    len1=stat_buff.st_size;
	}
      }
    }
    else
      infile=stdin;

    
    image.compression=compress;
    image.flags=old_flags;
    image.dpi_x=dpi_x;
    image.dpi_y=dpi_y;
    
    if (png||rtl||svg)
      if (land) image.flags|=LAND;
    read_image(&image,infile,user_trans,ctrans);
    if (!(png||rtl||svg))
      if (land) image.flags|=LAND;

    fclose(infile);
  
  
    if (image.components==0){
      c=1;
      if ((image.colour==A1GREY)||(image.colour==GREYA)) c=2;
      if (image.colour==RGB) c=3;
      if ((image.colour==A1RGB)||(image.colour==RGBA)) c=4;
      image.components=c;
    }


    if ((png)&&(overwrite)&&(infile!=stdin)){
      if ((strlen(file1)<4)||(strcasecmp(file1+strlen(file1)-4,".png")))
	error_exit("Cannot use -O with non-png input");
      file2=file1;
      rep_shrink=1;
    }
    if ((file2)&&(file2==file1)){
      outfile=fopen(file2,"wb"); /* need to avoid truncating file2 */
      if(!outfile){
	fprintf(stderr,"Error, unable to open %s for writing.\n",file2);
	exit(1);
      }
    }

    if (pdf) write_pdf(outfile,&image,bb,pt);
    else if (png) write_png(outfile,&image);
    else if (rtl) write_rtl(outfile,&image);
    else if (svg) write_svg(outfile,&image,svg-1);
    else write_ps(outfile,&image,bb,pt);

    if (overwrite) fclose(outfile);

    if (rep_shrink){
      if (!stat(file2,&stat_buff))
	len2=stat_buff.st_size;
      if ((len1)&&(len2))
	fprintf(stderr,"Original size: %d   final size: %d (%.1lf%%)\n",len1,
		len2,100*(double)len2/(double)len1);
    }

    free_image(&image);
    file1=argv[opt];
    opt++;
  }
  while((multi_image)&&(opt<argc+ofile+overwrite));

  /* Need to finalise PDF file */
  if (pdf) write_pdf(outfile,NULL,bb,pt);
  /* and PostScript files */
  if (ps) write_ps(outfile,NULL,bb,pt);
  
  if (!overwrite) fclose(outfile);
  
  return(0);

}

void free_image(struct bitmap *image){

  image->trans=NULL;
  image->dpi_x=image->dpi_y=0;
  image->bkgd=NULL;
  if (image->pal_len) free(image->palette);
  image->palette=NULL;
  if (image->apal_len) free(image->apalette);
  image->apalette=NULL;
  image->pal_len=0;
  image->apal_len=0;
  image->components=0;
  if (image->smask) {
    free(image->smask->data);
    free(image->smask);
  }
  image->smask=NULL;
  free(image->data);
  image->data=NULL;

}

void help(void){
  printf("usage:\n\nbmp2eps [options] [-format] [input [output]]\n\n");
  printf("  -format is [-pdf|-png|-svg|-html]\n\n" 
         " -an   An paper size (n=0 to 6) (implies -p)\n"
         " -b=   ignore background colour in image\n"
	 " -b=r,g,b use given colour as background for images with"
	 " transparency\n"
         " -d    include distillerparams in EPS\n"
         " -e    expand to fill page (implies -p)\n"
	 " -eps  produce EPS\n"
	 " -f    fast (poorer zlib compression)\n"
         " -html produce html output\n"
         " -i    use internal PNG reader, even if compiled with libpng\n"
         " -l    landscape\n");
  printf(" -n    use no compression\n"
	 " -o file  alternative to naming output as last file\n"
	 " -O    overwrite input file with output. Implies -png\n"
         " -p    centre on (default) page, and if EPS use PostScript instead\n"
	 " -pdf  produce PDF\n"
         " -png  produce PNG\n"
	 " -q    suppress informational messages\n");
  printf(" -r    RLE compress output"
#if DEFAULT_COMPRESS == RLE
	 " (default)"
#endif
	 "\n"
         " -rtl  produce RTL suitable for a DJ1050C and Z5600."
	 " Use -D to set size.\n"
         " -s    smooth output\n");
  printf(" -svg  produce SVG\n"
         " -t    ignore transparency\n"
         " -t=   make white transparent (short for -t=255,255,255)\n"
         " -t=r,g,b make given colour transparent\n"
         " -u    US letter (implies -p)\n"
         " -v    be verbose\n");
  printf(" -x    write palette in hex, not base 85\n"
#ifdef USE_ZLIB
	 " -z    zlib flate compress output"
#ifndef USE_ZOPFLI
	 " (default)"
#else
	 " (default is Zopfli flate)"
#endif
	 "\n"
#endif
         " -B=urx[,ury] specify bounding box\n"
         " -B=llx,lly,urx,ury specify bounding box\n"
         " -D    obey dpi setting from bmp/png/jpeg file\n"
         " -D=dpi use given dpi setting\n");
  printf(" -G3   use CCITT Group 3 1D encoding (on bilevel images)\n"
         " -G4   use CCITT Group 4 2D encoding (on bilevel images)\n"
	 " -N=d  lossy reduction to d bits per channel\n"
         " -P=n  select PNG predictors to try with flate/LZW compression\n"
         " -R    if possible, pass PNG data unmodified\n"
         " -V    list supported file types\n"
         " -X    write all data in hex\n"
#ifdef USE_LZW
         " -Z    LZW compress output"
#if DEFAULT_COMPRESS == LZW
         " (default)"
#endif
#endif
         "\n");
  printf(" -2    protect level 3 PostScript from level 2 interpreters\n"
	 " -7    use 7 bit output (default for PS/EPS)\n"
	 " -8    use 8 bit output (default for PDF)\n"
	 " -9    try harder to compress\n"
	 " -24   keep 24 bit images 24 bit,"
	 " even if fewer than 256 unique colours\n"
	 " -99   try yet harder to compress, repeat up to -999\n");
  printf("\nOutput is written to stdout and defaults to level 3 EPS.\n"
         "Use '-Z' (or some other compression) to produce level 2 EPS\n\n"
         "The =r,g,b colour specifiers for -b and -t are in decimal. They can "
         "also be\n"
	 "written =rrggbb in hex, and =g (decimal) for greyscale\n\n");
  printf("If the executable is named bmp2pdf, PDF is the default output type\n"
	 "                           bmp2png, PNG is the default output type\n");
  printf("\nOutput types of PDF and PS may contain multiple images\n");
  printf("PNG recompression with -O may specify multiple files to process\n");
  printf("To list input formats supported, specify -V\n");
  printf("\n               bmp2eps " BMP2EPS_VER " (c) 2002-2020 MJR\n"
         "               Distributed under the GPL v2\n\n");
  exit(0);
}

void versions(void){
  printf(" bmp2eps " BMP2EPS_VER " (c) 2002-2020 MJR\n\n"
         "The following input formats are supported:\n\n"
	 "BMP   MS Windows or OS/2 Bitmap with depth of 1, 4, 8, 24"
         " or 32 bits.\n"
         "      RLE compressed bitmap with depth of 4 or 8.\n"
	 "GIF   GIFs containing a single image.\n"
	 "JPEG  Colour or greyscale baseline JPEGs (not for "
#ifdef USE_ZLIB
	 "PNG or "
#endif
         "RTL output).\n");
  printf("PBM   Binary or ASCII data.\n"
         "PGM   Up to 8 bit greyscale, binary or ASCII data.\n"
	 "PNG   PNGs containing single image"
#ifdef USE_ZLIB
#ifdef USE_LIBPNG
         " (libpng available).\n"
#else
         " (internal reader only).\n"
#endif
#else
         " (limited support without zlib).\n"
#endif
	 "PPM   24 bit colour, binary or ASCII data.\n"
         "TGA   left to right, 8, 15, 16, 24 or 32 bits.\n\n");
  exit(0);
}
