/* Copyright (c) 2010 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* Examine an unpaletted 24- or 32-bit image, and count the unique colours.
 * If find >256, exit.
 * If find <=256, convert the image to paletted at a suitable depth
 *
 * Expect 32 bit image as RGBA. Store temporary palette as ARGB
 */

#include<stdio.h>
#include<stdlib.h>

#include "bmp2eps.h"

void colour_count(struct bitmap *image){
  int ncols,i,j,colour,debug,depth,tmp,greyscale,bytes_per_pixel;
  unsigned int size;
  unsigned int palette[257];
  unsigned char *p,*p2;
  static unsigned char buff;

  size=image->height*image->width;
  depth=image->depth;
  ncols=0;
  debug=image->flags&DEBUG_MASK;

  if ((image->depth!=16)&&(image->depth!=24)&&(image->depth!=32)){
    if (debug)
      fprintf(stderr,"Unexpected depth %d in colour_count\n",image->depth);
    return;
  }

  /* First check whether greyscale */
  if (image->depth>=24){
    bytes_per_pixel=3;
    if (image->depth==32) bytes_per_pixel=4;
    greyscale=1;
    p=image->data;
    for(i=0;i<size;i++){
      if ((*p!=*(p+1))||(*(p+1)!=*(p+2))){
	greyscale=0;
	break;
      }
      p+=bytes_per_pixel;
    }

    if (greyscale){
      if (debug)
	fprintf(stderr,"RGB data contains greyscale image. Reducing.\n");
      p=p2=image->data;
      if (image->depth==24){
	for(i=0;i<size;i++){
	  *(p++)=*p2;
	  p2+=3;
	}
	image->depth=8;
	image->data_len=size;
	image->data=realloc(image->data,image->data_len);
	image->colour=GREY;
	image->components=1;
	return;
      }
      else{
	p2+=2;
	for(i=0;i<size;i++){
	  *(p++)=*(p2++);
	  *(p++)=*p2;
	  p2+=3;
	}
	image->depth=16;
	image->data_len=2*size;
	image->data=realloc(image->data,image->data_len);
	image->colour=GREYA;
	image->components=2;
	if (image->bkgd) image->bkgd[0]=1;
      }
      /* Are now GREYA, so continue to process in case <=256 cols */
      depth=image->depth;
    }
  }
  
  p=image->data;
  for(i=0;i<size;i++){
    if (image->depth==16){
      colour=0x010101*(*p)+(((int)*(p+1))<<24);
      p+=2;
    }
    else{
      colour=(((int)*p)<<16)+(((int)*(p+1))<<8)+*(p+2);
      p+=3;
      if (depth==32) colour+=*(p++)<<24;
    }
    for(j=0;j<ncols;j++)
      if (colour==palette[j]) break;
    if (j==ncols){
      palette[ncols++]=colour;
      if (ncols==257) break;
    }
  }

  if ((ncols<257)&&(image->bkgd)){
    if (image->depth==16)
      colour=0x010101*image->bkgd[1]+0xff000000;
    else{
      colour=(image->bkgd[1]<<16)+(image->bkgd[2]<<8)+image->bkgd[3];
      if (depth==32) colour+=0xff000000; /* opaque */
    }
    for(j=0;j<ncols;j++)
      if (colour==palette[j]) break;
    if (j==ncols)
      palette[ncols++]=colour;
  }

  if (ncols==257){
    if (debug) fprintf(stderr,"More than 256 colours\n");
    return;
  }
  
  if (debug) fprintf(stderr,"%d colours: reducing to paletted\n",ncols);

  /* Want opaque colours at end of palette */

  if ((depth==16)||(depth==32)){
    i=0;
    j=ncols-1;
    while(i<j){
      while((palette[j]>>24)==0xff) j--;
      while ((i<j)&&((palette[i]>>24)!=0xff)) i++;
      if (i<j){
	tmp=palette[i];
	palette[i]=palette[j];
	palette[j]=tmp;
      }
    }
  }
  
  /* Reduce to paletted if possible */

  image->palette=malloc(3*ncols);
  if (!image->palette){
    fprintf(stderr,"Malloc error\n");
    exit(1);
  }

  for(i=0;i<ncols;i++){
    image->palette[3*i]=(palette[i]&0xff0000)>>16;
    image->palette[3*i+1]=(palette[i]&0xff00)>>8;
    image->palette[3*i+2]=(palette[i]&0xff);
  }

  if ((depth==16)||(depth==32)){
    if ((palette[0]>>24)!=0xff){
      if (((palette[0]>>24)==0)&&((palette[1]>>24)==0xff)){
	image->trans=&buff;
	*image->trans=0;
      }
      else{
	image->apalette=malloc(ncols);
	if (!image->apalette) error_exit("Malloc error for alpha palette");
	for(i=0;i<ncols;i++)
	  image->apalette[i]=palette[i]>>24;
	image->apal_len=ncols;
	while (image->apalette[image->apal_len-1]==0xff) image->apal_len--;
      }
    }
  }
  
  p=image->data;
  for(i=0;i<size;i++){
    if (image->depth==16){
      colour=0x010101*(*p)+(((int)*(p+1))<<24);
      p+=2;
    }
    else{
      colour=(*p<<16)+(*(p+1)<<8)+*(p+2);
      p+=3;
      if (depth==32) colour+=*(p++)<<24;
    }
    for(j=0;j<ncols;j++)
      if (colour==palette[j]) break;
    image->data[i]=j;
  }

  if ((depth==24)&&(image->trans)){
    colour=((image->trans[0])<<16)+((image->trans[1])<<8)+
           image->trans[2];
    for(j=0;j<ncols;j++)
      if (colour==palette[j]) break;
    if (colour!=palette[j]){
      fprintf(stderr,"Transparent colour %6x not found!\n",colour);
      image->trans=NULL;
    }
    *image->trans=j;
  }


  if (image->bkgd){
    if (image->depth==16)
      colour=0x010101*image->bkgd[1]+0xff000000;
    else{
      colour=(image->bkgd[1]<<16)+(image->bkgd[2]<<8)+image->bkgd[3];
      if (depth==32) colour+=0xff000000; /* opaque */
    }
    for(j=0;j<ncols;j++)
      if (colour==palette[j]) break;
    if (colour!=palette[j]){
      fprintf(stderr,"Background colour %6x not found!\n",colour);
      image->bkgd=NULL;
    }
    image->bkgd[0]=1;
    image->bkgd[1]=j;
  }

  p=realloc(image->data,size);
  if (p) image->data=p;

  image->data_len=size;
  image->depth=8;
  image->components=1;
  image->colour=PALETTE;
  image->type=PALETTE;
  image->pal_len=ncols;

  if (debug){
    fprintf(stderr,"Palette length %d entries",image->pal_len);
    if (image->apalette)
      fprintf(stderr," with alpha palette length %d",image->apal_len);
    fprintf(stderr,"\n");
  }
  
}
