/* Copyright (c) 2010 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* Reduce a bitmap from 8 bits per pixel to 1, 2 or 4 bits per pixel
 * in-place. Note that each row is rounded up to the nearest
 * integer number of bytes in length.
 *
 * Originally intended for paletted images, but should work with greyscale
 * if new depth passed in image->pal_len...
 */

#include<stdio.h>
#include<stdlib.h>

#include "bmp2eps.h"

void compact(struct bitmap *image){
  int ncols,i,j,height,width,debug,shift;
  unsigned char *p,*p2,ctmp;

  ncols=image->pal_len;
  height=image->height;
  width=image->width;
  debug=image->flags&DEBUG_MASK;

  if (ncols>16) return;

  if ((image->type==BMP)&&(image->colour==GREY)){ /* Need to move high bits */
    p=image->data;
    if (ncols==2)
      shift=7;
    else if (ncols<=4)
      shift=6;
    else
      shift=4;
    for(i=0;i<width*height;i++) p[i]>>=shift;
    if (image->trans) *(image->trans)>>=shift;
    if (image->bkgd) *(image->bkgd)>>=shift;
  }

  if (ncols==2){
    if (debug>1) fprintf(stderr,"Compacting to 1 bit / pixel\n");
    image->depth=1;
    p=p2=image->data;
    for(i=0;i<height;i++){
      for(j=0;j<width-7;j+=8,p2+=8)
        *p++=(*p2<<7)+(*(p2+1)<<6)+(*(p2+2)<<5)+(*(p2+3)<<4)
	  +(*(p2+4)<<3)+(*(p2+5)<<2)+(*(p2+6)<<1)+*(p2+7);
      if (j!=width){
        *p=0;
        for (;j<width;j++) *p=(*p<<1)+*p2++;
        *p=*p<<(8-(width&7));
        ctmp=*(p-1)&((1<<(8-(width&7)))-1); /* this might help the rle */
        *p|=ctmp;                               /* ditto */
        p++;
      }
    }
    image->data_len=height*((width+7)>>3);
  }
  else if (ncols<=4){
    if (debug>1) fprintf(stderr,"Compacting to 2 bits / pixel\n");
    image->depth=2;
    p=p2=image->data;
    for(i=0;i<height;i++){
      for(j=0;j<width-3;j+=4,p2+=4)
        *p++=(*p2<<6)+(*(p2+1)<<4)+(*(p2+2)<<2)+*(p2+3);
      if (j!=width){
        *p=0;
        for (;j<width;j++) *p=(*p<<2)+*p2++;
        *p=*p<<(2*(4-(width&3)));
        ctmp=*(p-1)&((1<<(2*(4-(width&3))))-1); /* this might help the rle */
        *p|=ctmp;                               /* ditto */
        p++;
      }
    }
    image->data_len=height*((width+3)>>2);
  }
  else if (ncols<=16){
    if (debug>1) fprintf(stderr,"Compacting to 4 bits / pixel\n");
    image->depth=4;
    p=p2=image->data;
    for(i=0;i<height;i++){
      for(j=0;j<(width-1);j+=2,p2+=2)
	*(p++)=(*p2<<4)+*(p2+1);
      if (j==(width-1)){
	*p=(*(p2++))<<4;
	*p|=*(p-1)&0xf;                       /* to help rle */
	p++;
      }
    }
    image->data_len=height*((width+1)>>1);
  }
  
  image->data=realloc(image->data,image->data_len);
}

/* Expand a bitmap from 1, 2 or 4 bits per pixel to 8 bits per pixel
 * in-place.
 *
 * Intended for paletted images and to reverse compact(),
 */

void decompact(struct bitmap *image){
  int ncols,i,j,k,kmax,debug,depth;
  int scanlen,pix_per_byte,mask,scale;
  unsigned char *p,*p2,ctmp;

  ncols=image->pal_len;
  scanlen=(image->width*image->depth+7)/8;
  debug=image->flags&DEBUG_MASK;

  depth=8;
  if (ncols<=16) depth=4;
  if (ncols<=4) depth=2;
  if (ncols==2) depth=1;

  if (depth==8) return;

  if (debug) fprintf(stderr,"Expanding from %d to 8 bits per pixel\n",depth);

  p=malloc(image->height*image->width);
  if (!p) error_exit("Malloc error expanding to one byte per pixel");
  p2=p;
  
  pix_per_byte=8/image->depth;
  mask=(1<<depth)-1;
  
  for(i=0;i<image->height;i++){
    kmax=pix_per_byte;
    for(j=0;j<scanlen;j++){
      if ((j+1)*pix_per_byte>image->width)
	kmax=image->width-j*pix_per_byte;
      ctmp=image->data[i*scanlen+j];
      for(k=0;k<kmax;k++)
	*(p2++)=(ctmp>>((pix_per_byte-k-1)*depth))&mask;
    }
  }

  if ((image->type==BMP)&&(image->colour==GREY)){
    scale=1;
    while (scale<256)
      scale=(scale<<depth)+1;
    scale=scale&0xff;
    for (i=0;i<image->height*image->width;i++)
      p[i]*=scale;
    if (image->trans) *(image->trans)*=scale;
    if (image->bkgd) *(image->bkgd)*=scale;
  }
  
  free(image->data);
  image->data=p;
  image->depth=8;
}
