/* Copyright (c) 2012-2020 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */

/* Apply PNG-style predictor to image, row by row.
 * No predictor used if found to be of no benefit
 *
 * Strictly one should compress the whole image with every combination
 * of rows predicted or unpredicted, i.e. 2^image->height tests.  This
 * is impractical. This code decides to predict or not based on
 * whether it makes the row smaller when compressed with a small
 * number of surrounding rows. This is wrong, as the previous data
 * will affect the compression ratio. It is slower, and arguably less
 * less wrong, than what libpng does...
 */

#if defined(USE_ZLIB) || defined (USE_LZW)
#include <stdio.h>
#include <stdlib.h>  /* malloc */
#include <string.h>  /* memcpy */

#include "bmp2eps.h"

#ifdef USE_ZLIB
#include <zlib.h>
static int compress2m(Bytef *dest, uLongf *destLen, const Bytef *source,
		      uLong sourceLen, int level, int memLevel,
		      int strategy);
int compress_final(unsigned char *data, unsigned char *out,
			  unsigned int data_len, int out_len,
			  int strategy);
#endif
#ifdef USE_LZW
unsigned int to_lzw(unsigned char *buff, unsigned char *out,
		    unsigned int buff_len, int len);
#endif
#ifdef USE_ZOPFLI
static unsigned int compress_zopfi(unsigned char *dest, unsigned int zlen,
				   unsigned char *src, unsigned int len);
#endif
  
static unsigned int squish(unsigned char *buff, unsigned char *out,
			   unsigned int buff_len, int len, int method);

static void predict(int method, unsigned char *source, unsigned char *dest,
		    unsigned int plen, int bits_per_pixel);
static void p_sub(unsigned char *source, unsigned char *dest,
                  unsigned int plen, int bytes_per_pixel);
static void p_up(unsigned char *source, unsigned char *dest, unsigned int plen,
		 int bytes_per_pixel);
static void p_av(unsigned char *source, unsigned char *dest, unsigned int plen,
		 int bytes_per_pixel);
static void p_paeth(unsigned char *source, unsigned char *dest,
                    unsigned int plen, int bytes_per_pixel);

void de_palette(struct bitmap *image);

/* variables visible to compress functions */
static int compress_level, squish_level, debug, flags;

static struct comp_param {
  int valid;
  int fixed;
  int zstrat;
  int ml;
  int level;
} cp;

void lzw_flate_set_level(int level){
  compress_level=level;
}

void lzw_flate(struct bitmap *image){
  unsigned char *org_data,*new_data,*p,*buffer;
  unsigned char *predicted_data,*predicted_compressed,*best_whole_compressed;
  unsigned char *predict_ptr,*zstart,*zend;
  int i,j,row_len,test1,test2,compress,method,best_method,*hits;
  unsigned int best,zsize,old_method;
  unsigned int *old_methods;
  int train,train_row,block,changes;
  int predictor[5];
  unsigned long zlen;
  int best_whole,best_method_whole;
  unsigned int best_size,org_data_len;
  struct bitmap alt_image;
  
  test1=0; /* Avoid spurious uninitialised use warning */
  best_whole_compressed=predicted_compressed=predicted_data=NULL;
  best_method=-2;
  best_whole=-1;
  cp.valid=0;
  cp.fixed=0;
  flags=image->flags;

  org_data=image->data;
  org_data_len=image->data_len;
  train=2048;
  if (image->compress_level>6) image->compress_level=6;

  compress_level=image->compress_level;
  train=train<<compress_level;
  squish_level=9;
  if (compress_level<0) squish_level=6;
  if (compress_level==0) squish_level=8;
  
  debug=image->flags&DEBUG_MASK;
  compress=image->compression;

  if ((compress!=FLATE)&&(compress!=LZW)&&(compress!=FLATEF))
    error_exit("Invalid call to lzw_flate");


#ifndef USE_ZLIB
  if ((compress==FLATE)||(compress=FLATEF)){
    fprintf(stderr,"Invalid call to lzw_flate\n"
            "FLATE requested but compiled without FLATE support\n");
    exit(1);
  }
#endif

#ifndef USE_LZW
  if (compress==LZW){
    fprintf(stderr,"Invalid call to lzw_flate\n"
            "LZW requested but compiled without LZW support\n");
    exit(1);
  }
#endif

  row_len=(image->width*image->depth+7)/8;

  if (row_len*image->height!=image->data_len){
    fprintf(stderr,"Inconsistency in LZW/Flate\n");
  }

  if ((compress==FLATE)||(compress==FLATEF))
    zlen=1.01*(image->data_len+image->height)+14;
  else
    zlen=(image->data_len+image->height)*3/2+3;

  new_data=malloc(zlen);
  if (!new_data) error_exit("Malloc error in LZW/Flate");
  buffer=NULL;
  test2=0; /* Avoid spurious warning about uninitialised use */

  if (debug) fprintf(stderr,"Trying %s compression",
		     compress==FLATE?"Flate":
		     (compress==FLATEF?"Flate zfiltered":"LZW"));

  /* Should we try prediction? */

  /* General view is yes for 8 bit greyscale or 24 bit colour, 
   * no for paletted. We'll try paletted and 4 bit greyscale too */

  if ((image->flags&PNG)||
      ((image->flags&PREDICT)&&(image->depth>=4)&&(image->type==BMP)&&
       (image->height>1))){

    if (debug) fprintf(stderr," with prediction\n");

    best_method_whole=0; /* suppress unnecessary gcc warning */

    /* If depth<8 the only sane predictor is Up, method 2 */

    if (image->depth<8){
      i=~0;
      i^=(1<<(P_SHIFT+1-1));
      i^=(1<<(P_SHIFT+3-1));
      i^=(1<<(P_SHIFT+4-1));
      image->flags&=i;
    }

    /* If paletted, av predictor is rarely good */


    if ((image->type==PALETTE)&&(compress_level<2))
      image->flags&=(~(1<<(P_SHIFT+3-1)));

    
    predicted_data=malloc(image->data_len+image->height);
    if (!predicted_data) error_exit("Malloc error for predicted data");

    predicted_compressed=malloc(zlen);
    if (!predicted_compressed)
      error_exit("Malloc error for predicted buffer");

    for (method=0;method<=4;method++){
      if ((method!=0)&&((image->flags&(1<<(P_SHIFT+method-1)))==0))
	continue;
      if (method==0)
	predict(0,image->data,predicted_data,image->width,image->depth);
      else
	predict(1,image->data,predicted_data,image->width,image->depth);
      for(i=1;i<image->height;i++)
	predict(method,image->data+i*row_len,predicted_data+i*(row_len+1),
		image->width,image->depth);
      zsize=squish(predicted_data,predicted_compressed,
		   image->height*(row_len+1),zlen,compress);
      if (method==0){
	best_whole=zsize;
	best_method_whole=0;
      }
      else if (zsize<best_whole){
	best_whole=zsize;
	best_method_whole=method;
      }
      if (debug>1) fprintf(stderr,
			   "Whole image prediction, predictor %d size %8d\n",
			   method,zsize);
    }

    if (best_method_whole!=4){ /* Need to regenerate predicted data */
      if (best_method_whole==0)
        predict(0,image->data,predicted_data,image->width,image->depth);
      else
	predict(1,image->data,predicted_data,image->width,image->depth);
      for(i=1;i<image->height;i++)
	predict(best_method_whole,image->data+i*row_len,
		predicted_data+i*(row_len+1),
		image->width,image->depth);
    }
    /* Is it (almost) certain that this will be discarded in favour of
     * unpredicted data? If so, use a cheap compression method.
     */
    if ((best_method_whole==0)&&(!(image->flags&PNG))&&(compress_level<1))
      best_whole=squish(predicted_data,predicted_compressed,
				image->height*(row_len+1),zlen,compress);
    else
      best_whole=compress_final(predicted_data,predicted_compressed,
				image->height*(row_len+1),zlen,compress);
    best_whole_compressed=malloc(best_whole);
    if (!best_whole_compressed) error_exit("Malloc error");
    memcpy(best_whole_compressed,predicted_compressed,best_whole);
    if (debug) fprintf(stderr,"Whole image, predictor %d, size %8d\n",
		       best_method_whole,best_whole);
    
    /* Zero counters for different predictor types */
    
    for(i=0;i<5;i++)
      predictor[i]=0;

    if (best_method_whole==0) predictor[0]=1;
    else predictor[1]=1;
    predictor[best_method_whole]+=image->height-1;

    train_row=train/row_len;
    if (train_row<1) train_row=1;
    if (train_row>image->height-1) train_row=image->height-1;
    block=train/row_len;
    if (block<=1) block=2;
    if (block>image->height) block=image->height;
    if (compress_level<=0) block=train_row=1;
    old_methods=malloc(block*sizeof(int));
    if (!old_methods) error_exit("Malloc error for old_methods");

    while(block>=2){
      if (debug) fprintf(stderr,"Pass block size=%d\n",block);
      for(i=1;i<image->height-block;i+=block/2){
	if (i>train_row)
	  zstart=predicted_data+(i-train_row)*(row_len+1);
	else
	  zstart=predicted_data;
	if (i<(image->height-train_row-1))
	  zend=predicted_data+(i+train_row+1)*(row_len+1)-1;
	else
	  zend=predicted_data+(image->height*(row_len+1)-1);

	test1=squish(zstart,predicted_compressed,(zend-zstart)+1,zlen,compress);
	best_method=-1;
	best_size=test1;
	for(j=0;j<block;j++)
	  old_methods[j]=predicted_data[(i+j)*(row_len+1)];
	for (method=0;method<=4;method++){
	  if ((method!=0)&&((image->flags&(1<<(P_SHIFT+method-1)))==0))
	    continue;
	  for (j=0;j<block;j++)
	    predict(method,image->data+(i+j)*row_len,
		    predicted_data+(i+j)*(row_len+1),
		    image->width,image->depth);
	  test2=squish(zstart,predicted_compressed,(zend-zstart)+1,zlen,
		       compress);
	  if (test2<best_size){
	    best_size=test2;
	    best_method=method;
	  }
	}
	if (best_method==-1){
	  for (j=0;j<block;j++)
	    predict(old_methods[j],image->data+(i+j)*row_len,
		    predicted_data+(i+j)*(row_len+1),
		    image->width,image->depth);
	}
	else{
	  for (j=0;j<block;j++)
	    predict(best_method,image->data+(i+j)*row_len,
		    predicted_data+(i+j)*(row_len+1),
		    image->width,image->depth);
	  for(j=0;j<block;j++)
	    predictor[old_methods[j]]--;
	  predictor[best_method]+=block;

	}
      }
      if (debug) {
	fprintf(stderr,"Predictors used: ");
	for(i=0;i<5;i++) fprintf(stderr,"%dx%d ",i,predictor[i]);
	fprintf(stderr,"\n");
      }
      block/=2;
    }
    free(old_methods);

    /* Final pass */
    buffer=malloc(zlen);
    if (!buffer) error_exit("malloc error for buffer in lzw_flate");
    if ((compress_level>0)||((image->data_len<1000000)&&(compress_level==0))){
      if (debug) fprintf(stderr,"Final pass\n");
      hits=malloc(image->height*sizeof(int));
      if (!hits) error_exit("malloc error for hits array in block_pass");
      for(i=0;i<image->height;i++) hits[i]=0;

      changes=0;
      best=0; /* suppress unnecessary gcc warning */
      for(i=image->height-1;i>=0;i--){
	predict_ptr=predicted_data+i*(row_len+1);
	old_method=*predict_ptr;
	if (i>train_row)
	  zstart=predicted_data+(i-train_row)*(row_len+1);
	else
	  zstart=predicted_data;
	if (i<(image->height-train_row-1))
	  zend=predicted_data+(i+train_row+1)*(row_len+1)-1;
	else
	  zend=predicted_data+(image->height*(row_len+1)-1);
	for(method=0;method<=4;method++){
	  if ((method!=0)&&((image->flags&(1<<(P_SHIFT+method-1)))==0))
	    continue;
	  if ((i==0)&&(method>1)) continue;
	  predict(method,image->data+i*row_len,predict_ptr,
		  image->width,image->depth);
	  zsize=squish(zstart,predicted_compressed,(zend-zstart)+1,zlen,compress);
	  if (method==0){
	    best=zsize;
	    best_method=0;
	  }
	  else if (zsize<best){
	    best=zsize;
	    best_method=method;
	  }
	  else if ((zsize==best)&&(hits[i]<2)){
	    best_method=method;
	  }
	}
	if (best_method!=4)
	  predict(best_method,image->data+i*row_len,predict_ptr,
		  image->width,image->depth);
	predictor[old_method]--;
	predictor[best_method]++;
	if (old_method!=best_method){
	  /*
	    fprintf(stderr,"%d: change %d to %d\n",i,old_method,best_method);
	  */
	  changes++;
	  hits[i]++;
	  if (hits[i]<5){
	    i=i+train_row+1;
	    if (i>image->height-1) i=image->height-1;
	  }
	}
	/*  fprintf(stderr,"%5d: %d\n",i,best_method); */
      }
      if (debug>2) fprintf(stderr,"Changes: %d\n",changes);
      free(hits);
    
      if (debug) {
	fprintf(stderr,"Predictors used: ");
	for(i=0;i<5;i++) fprintf(stderr,"%dx%d ",i,predictor[i]);
	fprintf(stderr,"\n");
      }

      if (((compress_level==0)&&(predictor[best_method]>0.95*image->height))||
	  ((compress_level==1)&&(predictor[best_method]>0.99*image->height))){
	test2=best_whole+1;
	if (debug) fprintf(stderr,"Too few changes to retest\n");
      }
      else{
	test2=compress_final(predicted_data,buffer,
			     image->data_len+image->height,zlen,compress);
	if (debug) fprintf(stderr,"Mixed compressed size:%d\n",test2);
      }
    } /* if ((compress_level)...) */
    else
      test2=best_whole+1;
    
    if (test2>best_whole){
      if (debug) fprintf(stderr,"Reverting to uniform method %d size %d\n",
			 best_method_whole,best_whole);
      test2=best_whole;
      memcpy(buffer,best_whole_compressed,best_whole);
      for(i=0;i<=4;i++)
	if (i==best_method_whole) predictor[i]=image->height;
	else predictor[i]=0;
    }


  /* buffer is compressed predicted image, length test2 
   * unless we have skipped prediction, in which case buffer=NULL */

    free(predicted_compressed);
    predicted_compressed=buffer;
    buffer=NULL;
    free(predicted_data);
    predicted_data=NULL;
    
  } /* End ((image->flags&PNG)|| ( ... (image->height>1)))*/
  else{ /* No prediction */
    if (debug) fprintf(stderr,"\n");
  }

  if (!(image->flags&PNG)){ /* We have the option of not predicting */
  
  /* We don't need the uncompressed predicted data any more,
   * so new_data is free to be (re-)used */
  
    test1=compress_final(image->data,new_data,image->data_len,zlen,FLATE);


  /* predicted_compressed is compressed predicted image, length test2
   * new_data is compressed unpredicted image, length test1 
   * test1 & test2 can be zero -- this means LZW was used,
   * and compression ineffective */

    if (debug>2){
      fprintf(stderr,"Unpredicted compressed length: %d\n",test1);
      if (buffer)
	fprintf(stderr,"Predicted compressed length:   %d\n",test2);
    }

    if ((buffer)&&(debug))
      fprintf(stderr,"Prediction saves %d bytes in compressed data\n",
	      test1-test2);

    if ((((test1==0)&&(test2==0))||
	 ((image->data_len-test1<40)&&
	  ((buffer==NULL)||(image->data_len-test2<140))))){
      if (debug) fprintf(stderr,"Neither prediction nor compression used\n");
      free(new_data);
      if (buffer) free(buffer);
      image->flags&=(~PREDICT);
      image->compression=0;
      return;
    }

    if ((predicted_compressed)&&(test1-test2>100)&&(test2>0)){
      image->data=predicted_compressed;
      image->data_len=test2;
      free(new_data);
      if (debug) fprintf(stderr,"Prediction used\n");
    }
    else{
      image->data=new_data;
      image->data_len=test1;
      if (predicted_compressed) free(predicted_compressed);
      image->flags&=(~PREDICT);
      if ((debug)&&(buffer)) fprintf(stderr,"Prediction not used\n");
    }

  }
  else{ /* Are PNG */
    free(new_data);
    image->data=predicted_compressed;
    image->data_len=test2;
  }

  if (debug) fprintf(stderr,"After compression %d bytes\n",image->data_len);
  if (best_whole_compressed) free(best_whole_compressed);
  
#ifdef USE_ZOPFLI
  if (((compress==FLATE)||(compress==FLATEF))&&(image->flags&ZOPFLI)&&
      (!(image->flags&PNG_RAW))){
  /* Final recompression with zopfli */

    zlen=(row_len+1)*image->height;
    p=malloc(zlen);
    if (!p) error_exit("Malloc error preparing for zopfli compression");
    i=uncompress(p,&zlen,image->data,image->data_len);
    if (i!=Z_OK) error_exit("Internal decompression error");
    free(image->data);
    buffer=malloc(1.01*zlen+14);
    if (!buffer) error_exit("Malloc error preparing for zopfli compression");
    if (debug) fprintf(stderr,"Recompressing with zopfli\n");
    zlen=squish(p,buffer,zlen,zlen*1.01+14,ZOPFLI);
    free(p);
    buffer=realloc(buffer,zlen);
    image->data=buffer;
    image->data_len=zlen;
    if (debug) fprintf(stderr,"After compression %d bytes\n",image->data_len);
  }
#endif

  if (image->smask){
    if (debug) fprintf(stderr,"Now compressing mask\n");
    lzw_flate(image->smask);
  }

  if (image->compression==FLATEF) image->compression=FLATE;

  if ((image->type==PALETTE)&&(image->compression=FLATE)){
    if ((image->pal_len>16)&&(image->data_len<6*image->pal_len+2*image->apal_len+20)){
      if (debug) fprintf(stderr,"Trying palette removal\n");
      alt_image=*image;
      alt_image.palette=malloc(3*image->pal_len);
      if (!alt_image.palette) error_exit("Malloc error for alt palette");
      memcpy(alt_image.palette,image->palette,3*image->pal_len);
      if (image->apalette){
	alt_image.apalette=malloc(image->apal_len);
	if (!alt_image.apalette) error_exit("Malloc error for alt apalette");
	memcpy(alt_image.apalette,image->apalette,image->apal_len);
      }
      alt_image.data=malloc(org_data_len);
      if (!alt_image.data) error_exit("Malloc error for alt data");
      memcpy(alt_image.data,org_data,org_data_len);
      alt_image.data_len=org_data_len;
      de_palette(&alt_image);
      lzw_flate(&alt_image);
      test1=image->data_len+3*image->pal_len+image->apal_len;
      if (debug){
	fprintf(stderr,"Image %d bytes with %d byte palette, total %d\n",
		image->data_len,3*image->pal_len+image->apal_len,test1);
	fprintf(stderr,"Depaletted image %d bytes\n",alt_image.data_len);
      }
      if (test1+10>alt_image.data_len){
	if (debug) fprintf(stderr,"Using depaletted image\n");
	free(image->data);
	free(image->palette);
	if (image->apalette) free(image->apalette);
	*image=alt_image;
      }
      else{
	free(alt_image.data);
      }
    }
  }

  if (org_data!=image->data) free(org_data);

  
  return;
}

static void predict(int method, unsigned char *source, unsigned char *dest,
		    unsigned int plen, int bits_per_pixel){
  int bytes_per_pixel,row_len;
  bytes_per_pixel=bits_per_pixel/8;
  row_len=(plen*bits_per_pixel+7)/8;
  if (bytes_per_pixel==0){
    bytes_per_pixel=1;
    plen=row_len;
  }
  switch (method){
  case 0:
    *dest++=0;
    memcpy(dest,source,row_len);
    return;
  case 1:
    p_sub(source,dest,plen,bytes_per_pixel);
    return;
  case 2:
    p_up(source,dest,plen,bytes_per_pixel);
    return;
  case 3:
    p_av(source,dest,plen,bytes_per_pixel);
    return;
  case 4:
    p_paeth(source,dest,plen,bytes_per_pixel);
    return;
  default:
    fprintf(stderr,"Invalid predictor in call to predict\n");
    exit(1);
  }
}

static void p_up(unsigned char *source, unsigned char *dest, unsigned int plen,
	  int bytes_per_pixel){
  unsigned char *s2;
  int i,blen;

  blen=plen*bytes_per_pixel;

  *dest++=2;  /* Code for "Up" predictor */
  s2=source-blen;

  for(i=0;i<blen;i++) *dest++=*source++-*s2++;

}

static void p_sub(unsigned char *source, unsigned char *dest, unsigned int plen,
	  int bytes_per_pixel){
  int i,blen;

  *dest++=1;  /* Code for "Sub" predictor */
  blen=plen*bytes_per_pixel;
  if (bytes_per_pixel==1){
    *dest++=*source;
    for(i=1;i<plen;i++) {
      *dest++=*(source+1)-*source;
      source++;
    }
  }
  else{
    for(i=0;i<bytes_per_pixel;i++)
      *dest++=*(source+i);
    for(i=bytes_per_pixel;i<blen;i++) {
      *dest++=*(source+bytes_per_pixel)-*source;
      source++;
    }
  }

}

/* predictor is average of left and above */
static void p_av(unsigned char *source, unsigned char *dest, unsigned int plen,
	  int bytes_per_pixel){
  int i,blen,av;
  unsigned char *s2;

  *dest++=3;  /* Code for "Av" predictor */
  blen=plen*bytes_per_pixel;
  s2=source-blen;

  if (bytes_per_pixel==1){
    *dest++=*source-*s2++/2;
    for(i=1;i<plen;i++) {
      av=((int)(*source)+(int)(*s2++))>>1;
      *dest++=*(source+1)-av;
      source++;
    }
  }
  else{
    for(i=0;i<bytes_per_pixel;i++)
      *dest++=*(source+i)-*s2++/2;
    for(i=bytes_per_pixel;i<blen;i++) {
      av=((int)(*source)+(int)(*s2++))>>1;
      *dest++=*(source+bytes_per_pixel)-av;
      source++;
    }
  }

}

static void p_paeth(unsigned char *source, unsigned char *dest,
                    unsigned int plen, int bytes_per_pixel){
  int i,blen,a,b,c,p,x;
  unsigned char *s2;

  *dest++=4;  /* Code for "Paeth" predictor */
  blen=plen*bytes_per_pixel;
  s2=source-blen;

  if (bytes_per_pixel==1){
    *dest++=*source-*s2++;
    for(i=1;i<plen;i++) {
      a=*source;
      b=*s2;
      c=*(s2-1);
      p=a+b-c;
      if ((abs(p-a)<=abs(p-b))&&(abs(p-a)<=abs(p-c))) x=a;
      else if (abs(p-b)<=abs(p-c)) x=b;
      else x=c;
      *dest++=*(source+1)-x;
      source++;
      s2++;
    }
  }
  else{
    for(i=0;i<bytes_per_pixel;i++)
      *dest++=*(source+i)-*s2++;
    for(i=bytes_per_pixel;i<blen;i++) {
      a=*source;
      b=*s2;
      c=*(s2-bytes_per_pixel);
      p=a+b-c;
      if ((abs(p-a)<=abs(p-b))&&(abs(p-a)<=abs(p-c))) x=a;
      else if (abs(p-b)<=abs(p-c)) x=b;
      else x=c;
      *dest++=*(source+bytes_per_pixel)-x;
      source++;
      s2++;
    }
  }

}

static unsigned int squish(unsigned char *data, unsigned char *out,
			   unsigned int data_len, int out_len, int method){

  int j;

#ifdef USE_ZLIB
  uLongf zlen;
#endif

  if (method==ZOPFLI){
#ifdef USE_ZOPFLI
      return compress_zopfi(out,out_len,data,data_len);
#else
      fprintf(stderr,"Zopfli requested, but unavailable\n");
#ifdef USE_ZLIB
      fprintf(stderr,"Falling back to zlib\n");
      method=FLATE;
#else
      exit(1);
#endif
#endif
  }
  
  if ((method==FLATE)||(method==FLATEF)){
#ifdef USE_ZLIB
      zlen=out_len;
      if (cp.valid)
	j=compress2m(out,&zlen,data,data_len,cp.level,cp.ml,cp.zstrat);
      else if (method==FLATE)
	j=compress2(out,&zlen,data,data_len,squish_level);
      else
	j=compress2m(out,&zlen,data,data_len,squish_level,9,Z_FILTERED);
      if(j){
	fprintf(stderr,"ZLIB returned error code %d\n",j);
	fprintf(stderr,"data_len=%d, out_len=%d\n",data_len,out_len);
	exit(1);
      }
      return zlen;
#else
      fprintf(stderr,"Flate compression requested, "
	      "but compiled without ZLIB\n");
      exit(1);
#endif
    }
    else{
#ifdef USE_LZW
      return to_lzw(data,out,out_len,data_len);
#else
      fprintf(stderr,"LZW compression requested, "
	      "but compiled without LZW\n");
      exit(1);
#endif
    }

}

#ifdef USE_ZLIB

/* MacOS, bless it */
#ifndef z_const
#define z_const
#endif


static int compress2m(Bytef *dest, uLongf *destLen, const Bytef *source,
		      uLong sourceLen, int level, int memLevel,
		      int strategy)
{
    z_stream stream;
    int err;

    stream.next_in = (z_const Bytef *)source;
    stream.avail_in = (uInt)sourceLen;

    stream.next_out = dest;
    stream.avail_out = (uInt)*destLen;
    if ((uLong)stream.avail_out != *destLen) return Z_BUF_ERROR;

    stream.zalloc = (alloc_func)0;
    stream.zfree = (free_func)0;
    stream.opaque = (voidpf)0;

    err = deflateInit2(&stream, level, Z_DEFLATED, 15, memLevel,
                       strategy);
    if (err != Z_OK) return err;

    err = deflate(&stream, Z_FINISH);
    if (err != Z_STREAM_END) {
        deflateEnd(&stream);
        return err == Z_OK ? Z_BUF_ERROR : err;
    }
    *destLen = stream.total_out;

    err = deflateEnd(&stream);
    return err;
}

int compress_final(unsigned char *data, unsigned char *out,
		   unsigned int data_len, int out_len,
		   int strategy){
  int j,ml,zstrat,filtered_winning;
  unsigned char *dat;
  uLongf zlen,l2,l3;

#ifdef USE_ZOPFLI
  if ((compress_level<=0)&&(flags&ZOPFLI))
    return squish(data,out,data_len,out_len,strategy);
#endif
  
  if (strategy==FLATEF){
    filtered_winning=0;
    zstrat=Z_FILTERED;
  }
  else if (strategy==FLATE){
    filtered_winning=1;
    zstrat=Z_DEFAULT_STRATEGY;
  }
  else
    return squish(data,out,data_len,out_len,strategy);
    
  ml=9;
  zlen=out_len;
  j=compress2m(out,&zlen,data,data_len,9,ml,zstrat);
  if (j!=Z_OK) error_exit("Unexpected zlib error");
  if (cp.fixed==0){
    cp.valid=1;
    cp.zstrat=zstrat;
    cp.ml=ml;
    cp.level=9;
  }
  if (compress_level<0) return zlen;

  if (debug>3) fprintf(stderr,"ml=%d l2=%ld\n",ml,zlen);
  dat=malloc(out_len);

  if (filtered_winning){
    l3=out_len;
    j=compress2m(dat,&l3,data,data_len,9,ml,Z_FILTERED);
    if (j!=Z_OK) error_exit("Unexpected zlib error");
    if (debug>3) fprintf(stderr,"Z_FILTERED ml=%d l2=%ld\n",ml,l3);
    if (l3<zlen){
      if (debug>3) fprintf(stderr,"Z_FILTERED ml=%d, %ld<%ld\n",ml,l3,zlen);
      memcpy(out,dat,l3);
      zlen=l3;
      if (cp.fixed==0){
	cp.zstrat=Z_FILTERED;
	cp.ml=ml;
	cp.level=9;
      }
      filtered_winning=2;
    }
    if (l3>zlen) filtered_winning=0;
  }
    
  
  /* Sometimes lower values of memLevel achieve better compression */
  
  for(ml=8;ml>1;ml--) {
    l2=out_len;
    j=compress2m(dat,&l2,data,data_len,9,ml,zstrat);
    if (j!=Z_OK) break;
    if (debug>3) fprintf(stderr,"ml=%d l2=%ld\n",ml,l2);
    if ((l2>zlen)&&(ml<7)&&(compress_level<1)) break;
    if (l2<zlen){
      if (debug>3) fprintf(stderr,"ml=%d, %ld<%ld\n",ml,l2,zlen);
      memcpy(out,dat,l2);
      zlen=l2;
      if (cp.fixed==0){
	cp.zstrat=zstrat;
	cp.ml=ml;
	cp.level=9;
      }
    }
    if (filtered_winning){
      l3=out_len;
      j=compress2m(dat,&l3,data,data_len,9,ml,Z_FILTERED);
      if (j!=Z_OK) error_exit("Unexpected zlib error");
      if (debug>3) fprintf(stderr,"Z_FILTERED ml=%d l2=%ld\n",ml,l3);
      if (l3<zlen){
	if (debug>3) fprintf(stderr,"Z_FILTERED ml=%d, %ld<%ld\n",ml,l3,zlen);
	memcpy(out,dat,l3);
	zlen=l3;
	if (cp.fixed==0){
	  cp.zstrat=Z_FILTERED;
	  cp.ml=ml;
	  cp.level=9;
	}
      }
      if (l3<l2) filtered_winning=2;
      if (l3>l2) filtered_winning--;
    }
  }

  /* And sometimes lower values of level. Try just one other here */
  if (strategy==FLATE) filtered_winning=1;
  for(ml=9;ml>1;ml--) {
    l2=out_len;
    j=compress2m(dat,&l2,data,data_len,3,ml,zstrat);
    if (j!=Z_OK) break;
    if (debug>3) fprintf(stderr,"l=3, ml=%d l2=%ld\n",ml,l2);
    if ((l2>zlen)&&(ml<8)&&(compress_level<1)) break;
    if ((l2>zlen)&&(ml<6)) break;
    if (l2<zlen){
      if (debug>3) fprintf(stderr,"l=3 ml=%d, %ld<%ld\n",ml,l2,zlen);
      memcpy(out,dat,l2);
      zlen=l2;
      if (cp.fixed==0){
	cp.zstrat=zstrat;
	cp.ml=ml;
	cp.level=3;
      }
    }
    if (filtered_winning){
      l3=out_len;
      j=compress2m(dat,&l3,data,data_len,3,ml,Z_FILTERED);
      if (j!=Z_OK) error_exit("Unexpected zlib error");
      if (debug>3) fprintf(stderr,"Z_FILTERED ml=%d l2=%ld\n",ml,l3);
      if (l3<zlen){
	if (debug>3) fprintf(stderr,"Z_FILTERED ml=%d, %ld<%ld\n",ml,l3,zlen);
	memcpy(out,dat,l3);
	zlen=l3;
	if (cp.fixed==0){
	  cp.zstrat=Z_FILTERED;
	  cp.ml=ml;
	  cp.level=3;
	}
      }
      if (l3<l2) filtered_winning=2;
      if (l3>l2) filtered_winning--;
    }
  }
  
  free(dat);
  return zlen;
}
			   

#ifdef USE_ZOPFLI

#include <zopfli/zopfli.h>

static unsigned int compress_zopfi(unsigned char *dest, unsigned int zlen,
				   unsigned char *src, unsigned int len){
  struct ZopfliOptions zopt;
  size_t out_len;
  unsigned char *out;
  int c;

  out=NULL;
  out_len=0;

  ZopfliInitOptions(&zopt);

  /* Default would be 15. We exceed this at -99 */
  c=4*(compress_level+1)+2*compress_level*compress_level;
  if (c<2) c=2;
  zopt.numiterations=c;
  
  ZopfliCompress(&zopt,ZOPFLI_FORMAT_ZLIB,src,len,&out,&out_len);

  if (out_len>zlen) error_exit("Too much data in compress_zopfli");

  memcpy(dest,out,out_len);
  free(out);

  return out_len;
}
  
#endif   /* USE_ZOPFLI */


#endif   /* USE_ZLIB */


#endif /* USE_ZLIB or USE_LZW */

