/* Copyright (c) 2010-2019 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* Check 8 bit depth paletted images to ensure all entries used, and
 * that palette is not grey-scale */

#include<stdio.h>
#include<stdlib.h>

#include "bmp2eps.h"

void palette_check(struct bitmap *image){
  int i,j,hit,debug;
  unsigned int size;
  unsigned char ncols;
  unsigned int *new_pal,col,*map;

  debug=image->flags&DEBUG_MASK;

  if (image->depth!=8) return;
  if (image->type!=PALETTE) return;
  
  /* Find highest used palette entry */

  ncols=0;
  size=image->data_len;

  for(i=0;i<size;i++) ncols=(ncols>image->data[i] ? ncols : image->data[i]);

  if (image->bkgd) ncols=(ncols>image->bkgd[1] ? ncols : image->bkgd[1]);
  
  if (ncols<(image->pal_len-1)){
    if (debug>1) fprintf(stderr,"Palette length reduced from %d to %d.\n",
	    image->pal_len,1+ncols);
    image->pal_len=1+ncols;
    if (image->apal_len>image->pal_len) image->apal_len=image->pal_len;
  }

  /* Trim opaque entries from top of alpha palette */
  
  if (image->apalette){
    if (image->apalette[image->apal_len-1]==255){
      if (debug>1) fprintf(stderr,"Alpha palette length reduced from %d ",
			   image->apal_len);
      i=image->apal_len;
      while((i)&&(image->apalette[i-1]==255)) i--;
      image->apal_len=i;
      if (debug>1) fprintf(stderr,"to %d\n",
			   image->apal_len);
      if (i==0){
	free(image->apalette);
	image->apalette=NULL;
      }
    }
  }	
  
  /* Check whether all palette entries are distinct */

  new_pal=malloc(image->pal_len*sizeof(int));
  map=malloc(image->pal_len*sizeof(int));
  if ((!new_pal)||(!map)){
    fprintf(stderr,"Malloc error in palette_check()\n");
    exit(1);
  }

  if ((image->trans)&&(image->apalette))
    fprintf(stderr,"Have both alpha and a transparent colour\n");
  
  size=0;
  for(i=0;i<image->pal_len;i++){
    col=(image->palette[3*i]<<16)+(image->palette[3*i+1]<<8)+
      image->palette[3*i+2];
    if (image->apalette){
      if (i<image->apal_len)
	col+=image->apalette[i]<<24;
      else
	col+=0xff000000;  /*opaque */
    }
    if ((image->trans)&&(i==image->trans[0]))
      col+=1<<24;
    hit=0;
    for(j=0;j<size;j++)
      if (col==new_pal[j]) {hit=1;map[i]=j;break;}
    if (!hit){
      new_pal[j]=col;
      map[i]=j;
      size++;
    }
  }
  if (size<image->pal_len){
    fprintf(stderr,"Duplicate palette entries. "
	    "Reducing from %d to %d entries\n",
	    image->pal_len,size);
    
    
  /* Construct new palette */
    for(i=0;i<size;i++){
      image->palette[3*i]=(new_pal[i]>>16)&0xff;
      image->palette[3*i+1]=(new_pal[i]>>8)&0xff;
      image->palette[3*i+2]=new_pal[i]&0xff;
      if ((image->apalette)&&(i<image->apal_len))
	image->apalette[i]=(new_pal[i]>>24)&0xff;
    }
    image->pal_len=size;
    if (image->apalette){
      if (debug) fprintf(stderr,"Old alpha palette length %d ",
			 image->apal_len);
      if (image->apal_len>size) image->apal_len=size;
      while (image->apalette[image->apal_len-1]==0xff) image->apal_len--;
      if (debug) fprintf(stderr,"new alpha palette length %d\n",
			 image->apal_len);
    }
    
    /* Update image data */
    size=image->data_len;
    for(i=0;i<size;i++)
      image->data[i]=map[image->data[i]];
  }

  if (image->bkgd) image->bkgd[1]=map[image->bkgd[1]];
  if (image->trans) image->trans[0]=map[image->trans[0]];
  
  free(map);
  free(new_pal);
}

void grey_check(struct bitmap *image){
  int i,j,grey,debug,depth,magic,test,ok,*used;
  unsigned int size, ncols;

  debug=image->flags&DEBUG_MASK;

  if (image->type!=PALETTE) return;
  if (image->apalette) return;
  
  ncols=image->pal_len;
  size=image->data_len;

  /* See if palette is greyscale */

  grey=1;
  for(i=0;i<image->pal_len;i++){
    if((image->palette[3*i]!=image->palette[3*i+1])||
       (image->palette[3*i]!=image->palette[3*i+2])){
      grey=0;
      break;
    }
  }

  if (!grey) return;

  if (debug) fprintf(stderr,"Image is greyscale\n");

  /* Worry about transparency. It is possible that we have the same
     shade of grey as both transparent and non-transparent */

  if (image->trans){
    ok=1;
    grey=image->palette[3*(*image->trans)];
    for(i=0;i<image->pal_len;i++){
      if (i==*image->trans) continue;
      if (image->palette[3*i]==grey) ok=0;
    }
    
    if (!ok){ /* Need to find an unused grey entry */
      if (debug) fprintf(stderr,
			 "Transparent entry is duplicate of"
			 " non-transparent entry. Changing.\n");
      if (ncols>16){
	used=malloc(256*sizeof(int));
	if (!used) error_exit("Malloc error in grey_check");
	for(i=0;i<256;i++) used[i]=0;
	for(i=0;i<image->pal_len;i++) used[image->palette[3*i]]=1;
	for(i=image->pal_len-1;i>=0;i--) if (used[i]==0) break;
	if (used[i]==1){
	  fprintf(stderr,"Unexpected transparency problem\n");
	  image->trans=NULL;
	}
        else{
          for(j=0;j<3;j++)
            image->palette[3*(*image->trans)+j]=i;
        }
      }
      else if (ncols>4){
	used=malloc(16*sizeof(int));
	if (!used) error_exit("Malloc error in grey_check");
	for(i=0;i<16;i++) used[i]=0;
	for(i=0;i<image->pal_len;i++) used[(image->palette[3*i]+8)>>4]=1;
	for(i=image->pal_len-1;i>=0;i--) if (used[i]==0) break;
	if (used[i]==1){
	  fprintf(stderr,"Unexpected transparency problem\n");
	  image->trans=NULL;
	}
        else{
          for(j=0;j<3;j++)
            image->palette[3*(*image->trans)+j]=i+(i<<4);
        }
      }
      else{
	used=malloc(4*sizeof(int));
	if (!used) error_exit("Malloc error in grey_check");
	for(i=0;i<4;i++) used[i]=0;
	for(i=0;i<image->pal_len;i++) used[(image->palette[3*i]+32)>>6]=1;
	for(i=image->pal_len-1;i>=0;i--) if (used[i]==0) break;
	if (used[i]==1){
	  fprintf(stderr,"Unexpected transparency problem\n");
	  image->trans=NULL;
	}
        else{
          for(j=0;j<3;j++)
            image->palette[3*(*image->trans)+j]=21*i+(i<<6);
        }
      }
      if ((debug>1)&&(image->trans))
        fprintf(stderr,"Was %x, is now %x\n", grey,
			   image->palette[3*(*image->trans)]);
    }	
  }

  
  if(ncols<=16){
    if (ncols==2){
      magic=255;
      depth=1;
    }
    else if (ncols<=4){
      magic=255/3;
      depth=2;
    }
    else{
      magic=255/15;
      depth=4;
    }
    test=1;
    for(i=0;i<ncols;i++){
      j=(image->palette[3*i]*magic)&255;
      if ((j!=0)&&(j!=1)&&(j!=255)){
	test=0;
	break;
      }
    }
    if (test){
      for(i=0;i<ncols;i++)image->palette[3*i]/=magic;
      image->depth=depth;
      if (debug>1) fprintf(stderr,"Greyscale depth is %d bit\n",depth);
    }else{    /* We have a <=4 bit image, but cannot use a 4 bit
	       * greyscale representation. Give up and use palette! */
      if (debug>1) fprintf(stderr,"ncols=%d, but not %d bit greyscale\n",
			 ncols,depth);
      return;
    }
  }

  
  /* See if palette is already ordered */

  for(i=0;i<image->pal_len;i++){
    if(image->palette[3*i]!=i){
      grey=0;
      break;
    }
  }

  if (!grey){  /* Need to order things */
    if (debug>2) fprintf(stderr,"Greyscale palette being reordered\n");
    for(i=0;i<size;i++) image->data[i]=image->palette[3*image->data[i]];
  }


  /* Worry about transparency */
  if (image->trans){
    /* To account for palette re-ordering */
    *image->trans=image->palette[*image->trans*3];
    /* To account for depth change */
    *image->trans=*image->trans&((1<<image->depth)-1);
  }

  if (image->bkgd){
    image->bkgd[1]=image->palette[3*image->bkgd[1]];
    /* To account for depth change */
    image->bkgd[1]=image->bkgd[1]&((1<<image->depth)-1);
  }
    
  
  /* Now we can ditch palette etc. */

  free(image->palette);
  image->palette=NULL;
  image->type=BMP;
  image->colour=GREY;
  image->pal_len=0;

  if (image->depth<=4){
      image->pal_len=1<<image->depth;
      compact(image);
      image->pal_len=0;
  }

  return;
}

int grey_bits(struct bitmap *image){
  int debug;
  unsigned size;
  unsigned char *p,*end;

  if ((image->type!=BMP)||(image->colour!=GREY)||(image->depth!=8))
    return -1;

  size=image->height*image->width;
  p=image->data;
  end=p+size;
  debug=image->flags&DEBUG_MASK;

  while((p<end)&&((*p==0)||(*p==0xff))) p++;

  if (p==end) return 1;

  while((p<end)&&((*p==0)||(*p==0x55)||(*p==0xaa)||(*p==0xff))) p++;

  if (p==end) return 2;

  while((p<end)&&(((*p)&0x0f)==(((*p)&0xf0)>>4))) p++;

  if (p==end) return 4;

  if (debug>2) fprintf(stderr,"Full 8 bit greyscale found %d offset %d\n",
     (int)(*p),(int)(p-image->data));

  return 8;
}

static int comp(const void *a, const void *b){
  unsigned int x,y;

  x=*((int*)a);
  y=*((int*)b);
  if (x==y) return 0;
  if (x<y) return -1;
  return 1;
}

void palette_sort(struct bitmap *image){
  unsigned int i,j,pal[256],map[256],colour,ok,debug;
  unsigned char *new_pal,*new_alpha;

  debug=image->flags&DEBUG_MASK;
  
  for(i=0;i<image->pal_len;i++)
    pal[i]=(((int)image->palette[3*i])<<16)+(((int)image->palette[3*i+1])<<8)+
      image->palette[3*i+2];

  if (image->apalette){
    for(i=0;i<image->apal_len;i++)
      pal[i]+=((unsigned int)image->apalette[i])<<24;
    for(i=image->apal_len;i<image->pal_len;i++)
      pal[i]+=0xff000000;
  }

  qsort(pal,image->pal_len,sizeof(int),comp);

  for(i=0;i<image->pal_len;i++){
    colour=(((int)image->palette[3*i])<<16)+(((int)image->palette[3*i+1])<<8)+
      image->palette[3*i+2];
    if (image->apalette){
      if (i<image->apal_len)
	colour+=((unsigned int)image->apalette[i])<<24;
      else
	colour+=0xff000000;
    }
    for(j=0;j<image->pal_len;j++){
      if(pal[j]==colour){
	map[i]=j;
	break;
      }
    }
  }

#if 0
  /* If multiple fully-transparent entries, remove them */
  if (image->apalette){
    i=0;
    while((i<image->apal_len)&&((pal[i]>>24)==0)) i++;
    i--;
    if (i>0){
      if (debug)
	fprintf(stderr,"Multiple fully-transparent palette entries found\n"
		"Removing %d entries\n",i);
      for(j=i+1;j<image->pal_len;j++)
	pal[j-i]=pal[j];
      for(j=0;j<image->pal_len;j++){
	if (map[j]>i) map[j]-=i;
	else map[j]=0;
      }
      image->pal_len-=i;
      image->apal_len-=i;
    }
  }
#endif

  ok=1;
  for(i=0;i<image->pal_len;i++){
    if (map[i]!=i){
      ok=0;
      break;
    }
  }

  if (ok) return;

  if (debug) fprintf(stderr,"Palette sorted\n");
  
  new_pal=malloc(3*image->pal_len);
  if (!new_pal) error_exit("Malloc error for new palette");
  if (image->apalette){
    new_alpha=malloc(image->apal_len);
    if (!new_alpha) error_exit("Malloc error for new alpha palette");
  }

  for(i=0;i<image->pal_len;i++){
    new_pal[3*i]=(pal[i]&0xff0000)>>16;
    new_pal[3*i+1]=(pal[i]&0xff00)>>8;
    new_pal[3*i+2]=pal[i]&0xff;
  }
  free(image->palette);
  image->palette=new_pal;
  
  if (image->apalette){
    for(i=0;i<image->apal_len;i++)
      new_alpha[i]=pal[i]>>24;
    if (new_alpha[image->apal_len-1]==0xff){
      if (debug) fprintf(stderr,"Reducing alpha palette length from %d",
			 image->apal_len);
      while((image->apal_len>0)&&
	    (new_alpha[image->apal_len-1]==0xff)) image->apal_len--;
      if (debug) fprintf(stderr," to %d\n",
			 image->apal_len);
    }
    free(image->apalette);
    image->apalette=new_alpha;
  }
  
  for(i=0;i<image->width*image->height;i++)
    image->data[i]=map[image->data[i]];

  if (image->bkgd) image->bkgd[1]=map[image->bkgd[1]];
  if (image->trans) image->trans[0]=map[image->trans[0]];

}
