/* Copyright (c) 2012-2020 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "bmp2eps.h"

unsigned to_group3(struct bitmap *image, unsigned char **out);
unsigned to_group4(struct bitmap *image, unsigned char **out);
void lzw_flate(struct bitmap *image);

void ppmread(FILE *in, struct bitmap *image, int binary);
void pgmread(FILE *in, struct bitmap *image, int binary);
void pbmread(FILE *in, struct bitmap *image, int binary);
void gifread(FILE *in, struct bitmap *image);
void jpegread(FILE *in, struct bitmap *image);
void bmpread(FILE *in, struct bitmap *image);
void pngread(FILE *in, struct bitmap *image);
void pngread_int(FILE *in, struct bitmap *image);
void tgaread(FILE *in, struct bitmap *image, unsigned char *buff);

void grey_check(struct bitmap *image);
int grey_bits(struct bitmap *image);
void palette_check(struct bitmap *image);
void palette_sort(struct bitmap *image);
void colour_count(struct bitmap *image);
void dealpha(struct bitmap *image);
void rgba2argb(struct bitmap *image);
void stripalpha(struct bitmap *image);
void stripalphapal(struct bitmap *image);
void rgba2smask(struct bitmap *image);
void alphapal2tcol(struct bitmap *image);

void bit_reduce(unsigned char *dat, unsigned int len, int depth);

void print_bitmap(struct bitmap *image);
void rotate_image(struct bitmap *image);

int read_image(struct bitmap *image, FILE *infile,
	       int user_trans, unsigned int ctrans[3]){

  int i,c,debug,compress;
#ifndef FIG2PDF
  unsigned char *imagez,*p;
#endif
  unsigned char buff[10];
  static unsigned char trans_buff[3];
  unsigned int width,height;
  int ps,pdf,rtl,png;
  
  debug=image->flags&DEBUG_MASK;

  /* Check magic */

  i=fread(buff,1,3,infile);
  if (i!=3){
      fprintf(stderr,"Unable to read from input.\n");
      perror(NULL);
      exit(1);
    }
  
  if ((buff[0]=='P')
      &&((buff[1]=='6')||(buff[1]=='3'))
      &&(strchr(" \t\r\n",buff[2]))) ppmread(infile,image,
                                             (buff[1]=='6')?1:0);
  else if ((buff[0]=='P')
      &&((buff[1]=='5')||(buff[1]=='2'))
      &&(strchr(" \t\r\n",buff[2]))) pgmread(infile,image,
                                             (buff[1]=='5')?1:0);
  else if ((buff[0]=='P')
      &&((buff[1]=='4')||(buff[1]=='1'))
      &&(strchr(" \t\r\n",buff[2]))) pbmread(infile,image,
                                             (buff[1]=='4')?1:0);
  else if ((buff[0]=='G')
           &&(buff[1]=='I')
           &&(buff[2]=='F')) gifread(infile,image);
  else if ((buff[0]==0xff)
           &&(buff[1]==0xd8)
           &&(buff[2]==0xff)) jpegread(infile,image);
  else if ((buff[0]==137)
           &&(buff[1]=='P')
           &&(buff[2]=='N')){
    if (image->flags&PNGI)
      pngread_int(infile,image);
    else
      pngread(infile,image);
  }
#ifndef FIG2PDF
  else if ((buff[0]=='B')
           &&(buff[1]=='M')) bmpread(infile,image);
  else if (((buff[1]&0xfe)==0)&&(buff[2]!=0)&&((buff[2]&0xf4)==0)) /* Maybe a TGA */
    tgaread(infile,image,buff);
#endif
  else {
    fprintf(stderr,"Error, unknown file type.\n");
    exit(1);
  }

  if (debug>1) fprintf(stderr,"Data read, %d bytes\n",image->data_len);

  width=image->width;
  height=image->height;
  compress=image->compression;

  if (debug>2) print_bitmap(image);
  else if (debug){
    fprintf(stderr,"Image %dx%d depth %d",width,height,
            image->depth);
    if (image->type==PALETTE){
      fprintf(stderr," with palette of %d entries",image->pal_len);
      if (image->apalette)
	fprintf(stderr," and %d alpha entries",image->apal_len);
    }
    else if (image->type==DCT)
      fprintf(stderr," with DCT compression");
    if (image->colour==GREY) fprintf(stderr," greyscale");
    fprintf(stderr,"\n");
    if (image->bkgd){
      fprintf(stderr,"Background colour %02x",image->bkgd[1]);
      if (image->bkgd[0]==3)
        fprintf(stderr,"%02x%02x",image->bkgd[2],image->bkgd[3]);
      fprintf(stderr,"\n");
    }
  }


/* Sort out transparency issues */

  switch (user_trans){
  case -1: /* User wants transparency ignored */
    if (image->trans) image->trans=NULL;
    break;
  case 1: /* User wants given colour to be transparent */
    if (image->type==PALETTE){
      for(i=0;i<image->pal_len;i++)
        if(((unsigned char)ctrans[0]==image->palette[3*i])&&
           ((unsigned char)ctrans[1]==image->palette[3*i+1])&&
           ((unsigned char)ctrans[2]==image->palette[3*i+2])){
           image->trans=trans_buff;
           *image->trans=(unsigned char)i;
           break;
        }
    }
    else{
      image->trans=trans_buff;
      image->trans[0]=(unsigned char)ctrans[0];
      image->trans[1]=(unsigned char)ctrans[1];
      image->trans[2]=(unsigned char)ctrans[2];
    }
    break;
  }

  if (compress==PNG){ /* Nothing to fiddle with */
    image->compression=FLATE;
    return 0;
  }

  if (compress==PNG_A7){ /* Nothing to fiddle with */
    return 0;
  }


  
  /* Alpha channel */

  if (image->flags&NO_BKGD) image->bkgd=NULL;

  if (image->components==0){
    c=1;
    if ((image->colour==A1GREY)||(image->colour==GREYA)) c=2;
    if (image->colour==RGB) c=3;
    if ((image->colour==A1RGB)||(image->colour==RGBA)) c=4;
    image->components=c;
  }

  ps=1;
  png=(image->flags&PNG)?1:0;
  pdf=(image->flags&PDF)?1:0;
  rtl=(image->flags&RTL)?1:0;
  if (png||pdf||rtl) ps=0;

  /* Remove alpha channel on request */
  if (((image->colour==RGBA)||(image->colour==GREYA))&&(image->flags&NO_TRANS))
    stripalpha(image);

  if (((image->colour==PALETTE)&&(image->apalette))&&(image->flags&NO_TRANS))
    stripalphapal(image);

  /* Always eliminate unnecessary alpha channel */
  if ((image->colour==RGBA)||(image->colour==GREYA))
    dealpha(image);

  /* PostScript needs RGBA reducing to A1RGB, GREYA to A1GREY */
  if (((image->colour==RGBA)||(image->colour==GREYA))&&(ps))
    rgba2argb(image);

  /* and palette with alpha reducing to single transparent entry */

  if (((image->colour==PALETTE)&&(image->apalette))&&(ps))
    alphapal2tcol(image);
  
  /* PDF needs RGBA reducing to RGB plus smask, GREYA to GREY plus smask,
     and palette with alpha to plain palette and smask */
  if (((image->colour==RGBA)||(image->colour==GREYA)||image->apalette)&&(pdf))
    rgba2smask(image);
  
  /* Unless PNG, need to eliminate Alpha */
  if (((image->colour==RGBA)||(image->colour==GREYA))&&(!png))
    stripalpha(image);

  
  if (image->reduce_depth){
    if (((image->depth==24)||(image->depth==32))
	&&(image->type==BMP)){
      if (debug) fprintf(stderr,"Lossy reduction to %d bits per channel\n",
			 image->reduce_depth);
      bit_reduce(image->data,image->data_len,image->reduce_depth);
      if (image->trans)
	bit_reduce(image->trans,image->depth/8,image->reduce_depth);
    }
    else if ((image->depth==8)&&(image->colour==GREY)&&(image->type==BMP)){
      if (debug) fprintf(stderr,"Lossy reduction to %d bits per channel\n",
			 image->reduce_depth);
      bit_reduce(image->data,image->data_len,image->reduce_depth);
      if (image->trans)
	bit_reduce(image->trans,1,image->reduce_depth);
    }
    else if (image->type==PALETTE){
      if (debug) fprintf(stderr,"Lossy reduction of palette to %d bits"
			 " per channel\n",image->reduce_depth);
      bit_reduce(image->palette,image->pal_len*3,image->reduce_depth);
      if (image->apalette)
	bit_reduce(image->apalette,image->apal_len,image->reduce_depth);
    }
    else{
      fprintf(stderr,"Unable to perform requested reduction to %d bits"
	      " per channel\n",image->reduce_depth);
    }
  }
  
  
  /* Check for trailing unused entries, duplicate entries, etc */

  if ((image->depth==8)&&(image->type==PALETTE)) palette_check(image);

  if ((image->depth==8)&&(image->type==PALETTE)) palette_sort(image);

  /* Check full colour image to ensure more than 256 colours */

  if ((image->depth==24)&&(image->type==BMP)&&
      (!(image->flags&KEEP24))) colour_count(image);

  /* If png can have alpha in palette */

  if ((png)&&(image->depth==32)&&(image->type==BMP)&&
      (!(image->flags&KEEP24))) colour_count(image);

  if ((png)&&(image->depth==16)&&(image->type==BMP)&&(image->colour==GREYA)&&
      (!(image->flags&KEEP24))) colour_count(image);

  /* Check paletted images for greyscale conversion */

  if ((image->depth==8)&&(image->type==PALETTE)) grey_check(image);

  /* Compact images with few palette entries */

  if ((image->type==PALETTE)&&(image->pal_len<=16)&&(image->depth==8))
      compact(image); 

  /* Try to reduce greyscale images in a similar fashion */

  if ((image->depth==8)&&(image->type==BMP)&&(image->colour==GREY)){
    i=grey_bits(image);
    if (debug>1) fprintf(stderr,"Greyscale bits used: %d\n",i);
    if ((i==1)||(i==2)||(i==4)){
      image->pal_len=1<<i;
      compact(image);
      image->pal_len=0;
    }
  }


  
  /* Rotation */

  if (image->flags&LAND){
    if ((image->compression!=DCT)&&(image->compression!=PNG)){
      rotate_image(image);
      image->flags&=(~LAND);
    }
  }
  
  /* Compression */

#ifndef FIG2PDF
  if ((compress==CCITT)||(compress==CCITTG4)){
    if (image->depth!=1){
      if (debug) fprintf(stderr,"CCITT compression requested, but image depth not 1\n"
                         "Reverting to default compression\n");
      compress=DEFAULT_COMPRESS;
    }
    else{
      if (compress==CCITT) i=to_group3(image,&p);
      else i=to_group4(image,&p);
      if (i){
        if (debug>1) fprintf(stderr,
                             "Before CCITT %d bytes\nAfter CCITT:  %d bytes\n",
                             image->data_len,i);
        free(image->data);
        image->data=p;
        image->data_len=i;
      }
      else compress=0;
    }
  }

  if (compress==RLE){
    if(image->depth==1){
      i=to_group4(image,&p);
      if (i){
        if (debug>1) fprintf(stderr,
                             "Before CCITT %d bytes\nAfter CCITT:  %d bytes\n",
                             image->data_len,i);
        free(image->data);
        image->data=p;
        image->data_len=i;
        compress=CCITTG4;
      }
      else compress=0;
    }
    else if (image->depth<=8){

      if (!(imagez=malloc(image->data_len+2+(image->data_len)/100))){
        fprintf(stderr,"Malloc error for RLE\n");
        exit(1);
      }

      i=to_rle(image->data,imagez,image->data_len);

      if (debug>1) fprintf(stderr,
                           "Before rle: %d bytes\nAfter rle:  %d bytes\n",
                           image->data_len,i);

      if(i<image->data_len){
        p=realloc(imagez,image->data_len);
        if (p) imagez=p;
        free(image->data);
        image->data=imagez;
        image->data_len=i;
      } else{
        compress=0;
        free(imagez);
      }
    }
    else if (image->depth==24) compress=RLE24;
    else{
      fprintf(stderr,"RLE specified with unexpected depth %d\n",image->depth);
      exit(1);
    }
  }
#endif

  image->compression=compress;

  if ((compress==LZW)||(compress==FLATE)||(compress==FLATEF)) lzw_flate(image);

  if (debug>2){
    fprintf(stderr,"At exit of read_image:\n");
     print_bitmap(image);
  }

  return 0;

}



void bit_reduce(unsigned char *dat, unsigned int len, int depth){
  int i,p,mask,offset;

  if (depth>=8) return;
  if (depth<=0) return;
  
  mask=(1<<depth)-1;
  mask=(mask<<(8-depth));

  if (depth>=4){
    for(i=0;i<len;i++){
      p=dat[i];
      p=(p&mask)+(p>>depth);
      dat[i]=p;
    }
  }
  else if (depth==3){
    for(i=0;i<len;i++){
      p=dat[i];
      p=(p&mask);
      offset=p>>depth;
      p=p+offset;
      offset=offset>>depth;
      p=p+offset;
      dat[i]=p;
    }
  }
  else if (depth==2){
    for(i=0;i<len;i++){
      p=dat[i];
      if (p<64) p=0;
      else if (p<128) p=0x55;
      else if (p<192) p=0xaa;
      else p=255;
      dat[i]=p;
    }
  }
  else if (depth==1){
    for(i=0;i<len;i++){
      p=dat[i];
      if (p<128) p=0;
      else p=255;
      dat[i]=p;
    }
  }

}

void de_palette(struct bitmap *image){
  int i,j,k,bytes_per_pixel,size,mask,shift,is_rgb;
  unsigned char *dat,*p,*pal,c;
  static unsigned char back[4],trans[3];

  if (!image->palette) return;

  is_rgb=0;
  pal=image->palette;
  for(i=0;i<image->pal_len;i++){
    if ((pal[3*i]!=pal[3*i+1])||(pal[3*i]!=pal[3*i+2])){
      is_rgb=1;
      break;
    }
  }

  if (image->trans){
    if ((image->trans[0]!=image->trans[1])||(image->trans[0]!=image->trans[2]))
      is_rgb=1;
  }
  
  if ((image->bkgd)&&(image->bkgd[0]==3)){
    if ((image->bkgd[1]!=image->bkgd[2])||(image->bkgd[1]!=image->bkgd[3]))
      is_rgb=1;
  }

  if (is_rgb)
    bytes_per_pixel=3;
  else
    bytes_per_pixel=1;
  if (image->apalette) bytes_per_pixel++;
  
  size=image->width*image->height;

  dat=malloc(size*bytes_per_pixel);
  if (!dat) error_exit("Malloc error in de_palette");

  p=dat;
  pal=image->palette;
  if (image->pal_len>16){
    for(i=0;i<size;i++){
      c=image->data[i];
      *(p++)=pal[3*c];
      if (is_rgb){
	*(p++)=pal[3*c+1];
	*(p++)=pal[3*c+2];
      }
      if (image->apalette){
	if (c<image->apal_len)
	  *(p++)=image->apalette[c];
	else
	  *(p++)=255;
      }
    }
  }
  else if (image->pal_len>4){
    for(i=0;i<image->height;i++){
      for(j=0;j<image->width;j+=2){
	c=image->data[j/2]>>4;
	*(p++)=pal[3*c];
	if (is_rgb){
	  *(p++)=pal[3*c+1];
	  *(p++)=pal[3*c+2];
	}
	if (image->apalette){
	  if (c<image->apal_len)
	    *(p++)=image->apalette[c];
	  else
	    *(p++)=255;
	}
	if (j+1<image->width){
	  c=image->data[j/2]&0xf;
	  *(p++)=pal[3*c];
	  if (is_rgb){
	    *(p++)=pal[3*c+1];
	    *(p++)=pal[3*c+2];
	  }
	  if (image->apalette){
	    if (c<image->apal_len)
	      *(p++)=image->apalette[c];
	    else
	      *(p++)=255;
	  }
	}
      }
    }
  }
  else if (image->pal_len>2){
    for(i=0;i<image->height;i++){
      for(j=0;j<image->width;j+=4){
	mask=0xc0;
	shift=6;
	for(k=0;k<4;k++){
	  if (j+k>image->width) break;
	  c=(image->data[j/4]&mask)>>shift;
	  *(p++)=pal[3*c];
	  if (is_rgb){
	    *(p++)=pal[3*c+1];
	    *(p++)=pal[3*c+2];
	  }
	  if (image->apalette){
	    if (c<image->apal_len)
	      *(p++)=image->apalette[c];
	    else
	      *(p++)=255;
	  }
	  mask=mask>>2;
	  shift-=2;
	}
      }
    }
  }
  else{
    for(i=0;i<image->height;i++){
      for(j=0;j<image->width;j+=8){
	mask=0x80;
	shift=7;
	for(k=0;k<8;k++){
	  if (j+k>image->width) break;
	  c=(image->data[j/8]&mask)>>shift;
	  *(p++)=pal[3*c];
	  if (is_rgb){
	    *(p++)=pal[3*c+1];
	    *(p++)=pal[3*c+2];
	  }
	  if (image->apalette){
	    if (c<image->apal_len)
	      *(p++)=image->apalette[c];
	    else
	      *(p++)=255;
	  }
	  mask=mask>>1;
	  shift-=1;
	}
      }
    }
  }

  if (image->bkgd){
    back[0]=1;
    if (is_rgb) back[0]=3;
    back[1]=pal[3*image->bkgd[1]];
    back[2]=pal[3*image->bkgd[1]+1];
    back[3]=pal[3*image->bkgd[1]+2];
    image->bkgd=back;
  }
  
  if (image->trans){
    trans[0]=pal[3*image->trans[0]];
    trans[1]=pal[3*image->trans[0]+1];
    trans[2]=pal[3*image->trans[0]+2];
    image->trans=trans;
  }
  
  if (image->apalette){
    free(image->apalette);
    image->apalette=NULL;
    image->apal_len=0;
  }
  free(image->palette);
  image->palette=NULL;
  image->pal_len=0;
  image->type=BMP;
  free(image->data);
  image->data=dat;
  image->data_len=size*bytes_per_pixel;

  
  if (bytes_per_pixel==1){
    image->colour=GREY;
    image->components=1;
    image->depth=8;
  }
  else if (bytes_per_pixel==2){
    image->colour=GREYA;
    image->components=2;
    image->depth=16;
  }
  else if (bytes_per_pixel==3){
    image->colour=RGB;
    image->components=3;
    image->depth=24;
  }
  else if (bytes_per_pixel==4){
    image->colour=RGBA;
    image->components=4;
    image->depth=32;
  }
    
  
}
