/* Copyright (c) 2010, 2013 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

#include <stdio.h>
#include <stdlib.h>

/* This routine is intended to read:
 *
 * MS Windows v3 bitmaps (Win 3.0, 3.1, NT3.5)
 * OS/2 1.x bitmaps
 * OS/2 2.x and 3 bitmaps
 *
 * it might also read MS Windows V4 and V5, however these are rarely seen
 *  in files.
 *
 * OS/2 1.x has three bytes per palette entry, not four, and
 *  no dpi information
 *
 * The mask information immediately follows the header if the header is
 * 40 bytes long. In larger headers, its offset is the same, but it counts
 * as being part of the header.
 */

#include "bmp2eps.h"
void de_msrle(struct bitmap *image, unsigned char *in, unsigned len);
void bmp16read(FILE *in, unsigned char *header, struct bitmap *image);

void bmpread(FILE *in, struct bitmap *image){
  unsigned char buff[60],encode,*p,*q,c;
  unsigned int *ip,itmp,red_mask,green_mask,blue_mask;
  unsigned int red_shift,green_shift,blue_shift;
  unsigned width,height,depth,dw,offset,off2,size,excess;
  int i,j,ncols,debug,os2,rle,bgr;

  /* avoid spurious uninitialised variable warnings */

  red_mask=green_mask=blue_mask=red_shift=green_shift=blue_shift=0;

  /* os2 is set to one only if os2 1.x with short palette */

  debug=image->flags&DEBUG_MASK;
  if (debug) fprintf(stderr,"BMP file detected\n");

  off2=3;
  rle=size=ncols=0;

  /* Read rest of first header */
  fread(buff,11,1,in);
  off2+=11;

  if ((buff[3])||(buff[4])||(buff[5])||(buff[6])){
    fprintf(stderr,"Not a valid BMP file\n");
    exit(1);
  }

  offset=buff[7]+(buff[8]<<8)+(buff[9]<<16)+(buff[10]<<24);
  if (debug>2) fprintf(stderr,"Offset to data in BMP: %d\n",offset);

  /* Read info header */

  fread(buff,12,1,in);
  off2+=12;
  encode=0;
  excess=0;

  if ((buff[0]==12)&&(buff[1]==0)){ /* OS/2 1.x style BMP */
    os2=1;
    if ((buff[8]!=1)||(buff[9])){
      fprintf(stderr,"Not a valid OS/2 BMP file\n");
      exit(1);
    }

    width=buff[4]+((unsigned)buff[5]<<8);
    height=buff[6]+((unsigned)buff[7]<<8);
    depth=buff[10];
  }
  else{ /* Must be Windows style BMP */
    os2=0;
    /* Known valid header lengths at this point are 40, 64, 108 and 124 */
    /* So accept from 40 to 255 */
    if ((buff[0]<40)||(buff[1])||(buff[2])||(buff[3])){
      fprintf(stderr,"Not a valid BMP file\n");
      exit(1);
    }
    if (debug>1) fprintf(stderr,"BMP header length: %d\n",(int)buff[0]);
    excess=buff[0]-40;

    fread(buff+12,40-12,1,in);
    off2+=40-12;

    width=buff[4]+((unsigned)buff[5]<<8)+((unsigned)buff[6]<<16)
      +((unsigned)buff[7]<<24);
    height=buff[8]+((unsigned)buff[9]<<8)+((unsigned)buff[10]<<16)
      +((unsigned)buff[11]<<24);

    if ((buff[12]!=1)||(buff[13])||(buff[15])){
      fprintf(stderr,"Not a valid BMP file\n");
      exit(1);
    }

    depth=buff[14];
    encode=buff[16];
    size=buff[20]+(((int)(buff[21]))<<8)+(((int)(buff[22]))<<16)+
      (((int)(buff[23]))<<24);

    if (image->dpi_x==-1){
      image->dpi_x=0.0254*(buff[24]+(((int)(buff[25]))<<8)+
			   (((int)(buff[26]))<<16)+(((int)(buff[27]))<<24));
      image->dpi_y=0.0254*(buff[28]+(((int)(buff[29]))<<8)+
			   (((int)(buff[30]))<<16)+(((int)(buff[31]))<<24));
    }
  }

  if (debug) fprintf(stderr,"Size %dx%d depth %d bits\n",width,height,depth);

  if (encode!=0){
    if ((encode==1)&&(depth==8)) rle=8;
    else if ((encode==2)&&(depth==4)) rle=4;
    else if (encode!=3){
      fprintf(stderr,"Encoding not supported\n");
      exit(1);
    }
  }

  if ((depth!=1)&&(depth!=4)&&(depth!=8)&&(depth!=16)&&(depth!=24)
      &&(depth!=32)){
    fprintf(stderr,"Depth of %d not supported\n",depth);
    exit(1);
  }

  bgr=1;
  if (encode==3){ /* Must read 12 byte pseudopalette */
    if (debug>1) fprintf(stderr,"Bitfield image\n");
    fread(buff+40,12,1,in);
    if (excess>=12) excess-=12;  /* If excess>12, this field will have been */
    off2+=12;                    /* counted within the header, if not, not */ 
    if ((depth==24)||(depth==32)){
      red_mask=*(unsigned int*)(buff+40);
      green_mask=*(unsigned int*)(buff+44);
      blue_mask=*(unsigned int*)(buff+48);
      if (debug>1)
        fprintf(stderr,"RGB masks: %x %x %x\n",red_mask,green_mask,blue_mask);
      if ((!red_mask)||(!green_mask)||(!blue_mask)){
	fprintf(stderr,"Error, invalid masks in bitfield image\n");
	exit(1);
      }
      red_shift=0;
      green_shift=0;
      blue_shift=0;
      while((red_mask&(1<<red_shift))==0) red_shift++;
      while((green_mask&(1<<green_shift))==0) green_shift++;
      while((blue_mask&(1<<blue_shift))==0) blue_shift++;
    }
  }

  /* Consume unwanted bits of header */

  if (excess>0) {
    off2+=excess;
    for (;excess;excess--) fgetc(in); /* Can't seek stdin */
  }


  if(depth<=8){ /* We have a palette to read */

    if (!os2) ncols=buff[32]+(buff[33]<<8)+(buff[34]<<16)+(buff[35]<<24);
    if (!ncols) ncols=1<<depth;
      
    if (ncols>(1<<depth)){
      fprintf(stderr,"Inconsistent palette length of %d.\n",ncols);
      exit(1);
    }

    image->pal_len=ncols;
    image->palette=malloc(3*image->pal_len);
    off2+=(4-os2)*image->pal_len;
    if (!image->palette){fprintf(stderr,"Malloc error\n");exit(1);}
    p=image->palette;
    for(i=0;i<image->pal_len;i+=2){
      fread(buff,8-2*os2,1,in);  /* MS Win uses BGR0, OS/2 uses BGR */
      *(p++)=buff[2];
      *(p++)=buff[1];
      *(p++)=buff[0];
      *(p++)=buff[6-os2];
      *(p++)=buff[5-os2];
      *(p++)=buff[4-os2];
    }
    image->type=PALETTE;
    image->colour=PALETTE;
  }else{
    image->type=BMP;
    image->colour=RGB;
  }


  if (offset<off2){
    fprintf(stderr,"Error in BMP format. %d %d\n",offset,off2);
    exit(1);
  }
  else if (offset>off2) {
    if (debug) fprintf(stderr,"%d bytes of extra data\n",offset-off2);
    for (;off2<offset;off2++) fgetc(in);
  }

  image->height=height;
  image->width=width;

  dw=(width*depth+7)/8;
  image->data_len=dw*height;
  image->depth=depth;

  if (depth==16){
    bmp16read(in, buff, image);
    return;
  }

  if(!rle){
    image->data=malloc(dw*height);
    if (!image->data){fprintf(stderr,"Malloc error\n");exit(1);}
    for (i=1;i<=height;i++){
      fread(image->data+(height-i)*dw,dw,1,in);
      if (dw&3) fread(buff,4-(dw&3),1,in); /* Lines are dword aligned */
    }
    
    if(depth==32){
      if (encode==3){
	ip=(unsigned int*)image->data;
	q=image->data;
        for(i=0;i<height;i++){
	  for(j=0;j<width;j++){ /* Never padding bytes at this depth */
            itmp=*(ip++);
	    *(q++)=(itmp&red_mask)>>red_shift;
	    *(q++)=(itmp&green_mask)>>green_shift;
	    *(q++)=(itmp&blue_mask)>>blue_shift;
	  }
	}
        bgr=0;
      }
      else{
        p=image->data;
        q=image->data;
        for(i=0;i<height;i++){
	  for(j=0;j<width;j++){ /* Never padding bytes at this depth */
	    *(q++)=*(p++);
	    *(q++)=*(p++);
	    *(q++)=*(p++);
	    p++;
	  }
	}
      }
      depth=24;
      dw=(width*depth+7)/8;
      image->data_len=dw*height;
      image->depth=depth;
    }

    if(bgr&&(depth==24)){  /* BMP is BGR not RGB ... */
      p=image->data;
      for(i=0;i<height;i++){
	for(j=0;j<width;j++){
	  c=*p;
	  *p=*(p+2);
	  *(p+2)=c;
	  p+=3;
	}
      }
    }
  }else{  /* if (rle) */
    p=malloc(size);
    if (!p){fprintf(stderr,"Malloc error\n");exit(1);}
    fread(p,size,1,in);
    de_msrle(image,p,size);
    free(p);
  }
  
}

void bmp16read(FILE *in, unsigned char *header, struct bitmap *image){
  unsigned char *palette;
  unsigned char *rowptr,*ptr;
  unsigned int width,height,colour;
  int is15,i,j;

  is15=1;
  width=image->width;
  height=image->height;

  if (header[16]==3){ /* Read "palette" and check correct
		       * Palette can be 7c00 03e0 001f (15 bit)
		       * or             f800 07e0 001f (16 bit)
		       * as represents RGB bitmasks. However,
		       * Intel byte ordering afflicts the above,
                       * and the entries are stored as dwords,
                       * not words. */
    palette=header+40;
    if ((palette[0]!=0)||(palette[4]!=0xe0)|| /* could check more than this */
	(palette[8]!=0x1f)||(palette[9]!=0)){
      fprintf(stderr,"Unexpected palette in 16 bit bmp. Exiting.\n");
      exit(1);
    }
    if (palette[1]&0x80) is15=0;
  }else if (header[30]!=0){
    fprintf(stderr,"Unexpected compression type %d for 16 bit bmp."
	    " Exiting.\n",(int)header[30]);
    exit(1);
  }

  if (image->flags>1){
    if (is15) fprintf(stderr,"15 bit 555\n");
    else fprintf(stderr,"16 bit 565\n");
  }

  image->depth=24;
  image->data_len=3*width*height;
  image->data=malloc(3*width*height);
  if (!image->data){fprintf(stderr,"Malloc error\n");exit(1);}

  rowptr=malloc(2*width);
  if (!rowptr){fprintf(stderr,"Malloc error\n");exit(1);}

  for(i=1;i<=height;i++){
    fread(rowptr,2*width,1,in);
    ptr=image->data+(height-i)*3*width;
    for(j=0;j<width;j++){
      colour=rowptr[2*j]+(rowptr[2*j+1]<<8);
      *(ptr+2)=(colour&0x1f)<<3; /* blue */
      if (is15){                 /* green */
	*(ptr+1)=(colour&0x03e0)>>2;
      }else{
	*(ptr+1)=(colour&0x07e0)>>3;
	colour>>=1;
      }
      *ptr=(colour&0x7c00)>>7;  /* red */
      ptr+=3;
    }
    if (width&1) fread(rowptr,2,1,in); /* rows are dword aligned */
  }
  free(rowptr);
}
