/* Copyright (c) 2013-2019 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* memmove */
#ifdef USE_ZLIB
#include <zlib.h>
#endif
#include "bmp2eps.h"

static void de_up(unsigned char *source, unsigned char *dest, 
		  unsigned char *last, unsigned int plen,
		  int bytes_per_pixel);

static void de_av(unsigned char *source, unsigned char *dest, 
		  unsigned char *last, unsigned int plen,
		  int bytes_per_pixel);

static void de_paeth(unsigned char *source, unsigned char *dest, 
		     unsigned char *last, unsigned int plen,
		     int bytes_per_pixel);

static void de_sub(unsigned char *source, unsigned char *dest, 
		   unsigned int plen, int bytes_per_pixel);

static void de_predict(struct bitmap *image, unsigned char *dat,
		       int channel_depth);

static void de_adam7(struct bitmap *image, unsigned char *dat,
		       int channel_depth);

static void de_adam7_bits(struct bitmap *image, unsigned char *dat,
			  int channel_depth);

void pngread_int(FILE *in, struct bitmap *image){
  int debug,i,j,hit;
  unsigned alpha,len,ilen,channel_depth,adam7;
  static unsigned char index[3];
  unsigned char buff[32],*iptr,*back;
  static unsigned char bbuff[4];
#ifdef USE_ZLIB
  unsigned char *p,*d,*dat;
  unsigned bytes_per_pixel;
  uLongf zlen;
#endif
  
  debug=image->flags&DEBUG_MASK;
  back=NULL;

  if (debug) fprintf(stderr,"PNG file detected\n");
  if (debug>2) fprintf(stderr,"Not using libpng\n");

  /* Three bytes consumed before this routine was called, now
     read rest of 8-byte magic */

  fread(buff,5,1,in);

  /* Now read first chunk, which must be an IHDR chunk */

  fread(buff,25,1,in);

  if((buff[4]!='I')||(buff[5]!='H')||(buff[6]!='D')||(buff[7]!='R')){
    fprintf(stderr,"Failed to locate IHDR chunk in readpng.\n");
    exit(1);
  }

  if(buff[3]!=13){
    fprintf(stderr,"Unexpected length for IHDR chunk in readpng.\n");
    exit(1);
  }

  image->width=buff[11]+(buff[10]<<8)+(buff[9]<<16)+(buff[8]<<24);
  image->height=buff[15]+(buff[14]<<8)+(buff[13]<<16)+(buff[12]<<24);

  alpha=0;
  image->type=BMP;
  image->components=1;

  if(buff[17]==0)
    image->colour=GREY;
  else if (buff[17]==2)
    {image->colour=RGB;image->components=3;}
  else if (buff[17]==3)
    image->colour=image->type=PALETTE;
  else if (buff[17]==4)
    {image->colour=GREYA; alpha=1;image->components=2;}
  else if (buff[17]==6)
    {image->colour=RGBA; alpha=1;image->components=4;}
  else{
    fprintf(stderr,"Unexpected image type %d in readpng. Aborting.\n",
	    (int)buff[17]);
    exit(1);
  }

  image->depth=buff[16]*image->components;
  channel_depth=buff[16];
  if (debug>1) fprintf(stderr,"Image %dx%d depth %d %s\n",image->width,
		     image->height,image->depth,alpha?"(with alpha)":"");

  if (buff[18]!=0){
    fprintf(stderr,"Unexpected compression type %d in readpng. Aborting.\n",
	    (int)buff[18]);
    exit(1);
  }

  if (buff[19]!=0){
    fprintf(stderr,"Unexpected filter type %d in readpng. Aborting.\n",
      (int)buff[19]);
    exit(1);
  }

  if (buff[20]>1){
    fprintf(stderr,"Unsupported interlace type %d in readpng. Aborting.\n",
      (int)buff[20]);
    exit(1);
  }
  adam7=buff[20];


  /* Loop through other chunks */

  iptr=NULL;
  ilen=0;
  while (1){
    if (fread(buff,8,1,in)!=1){
      fprintf(stderr,"Read error\n");
      exit(1);
    }
    len=buff[3]+(buff[2]<<8)+(buff[1]<<16)+(buff[0]<<24);
    /* PLTE */
    if ((buff[4]=='P')&&(buff[5]=='L')&&(buff[6]=='T')&&(buff[7]=='E')){
      if (image->colour!=PALETTE){
	fprintf(stderr,"Error in readpng: palette in non-paletted image.\n");
	exit(1);
      }
      image->pal_len=len/3;
      if (debug>1) fprintf(stderr,"Reading palette of %d entries.\n",
			   image->pal_len);
      image->palette=malloc(len);
      if (!image->palette){
	fprintf(stderr,"Malloc error for palette.\n");
	exit(1);
      }
      fread(image->palette,1,len,in);
    }
    /* pHYs */
    else if ((buff[4]=='p')&&(buff[5]=='H')&&(buff[6]=='Y')&&(buff[7]=='s')){
      if(len!=9){
	fprintf(stderr,"Invalid length for pHYs field.\n");
	exit(1);
      }
      fread(buff+8,1,len,in);
      if (image->dpi_x!=-1){
        if (debug>1) fprintf(stderr,"Ignoring DPI information, -D to obey\n");
      }
      else
        if (buff[16]==1){
          i=buff[11]+(buff[10]<<8)+(buff[9]<<16)+(buff[8]<<24);
          image->dpi_x=i*0.0254+.4;
          i=buff[15]+(buff[14]<<8)+(buff[13]<<16)+(buff[12]<<24);
          image->dpi_y=i*0.0254+.4;
          if (debug>1) fprintf(stderr,"DPI: %dx%d\n",image->dpi_x,image->dpi_y);
        }
        else{
          fprintf(stderr,"Ignoring DPI information, unknown units %d",
                  buff[16]);
        }
    }
    /* bKGD */
    else if ((buff[4]=='b')&&(buff[5]=='K')&&(buff[6]=='G')&&(buff[7]=='D')){
      if (len>6){
        fprintf(stderr,"Invalid length for bKGD field.\n");
        exit(1);
      }
      if (debug>1) fprintf(stderr,"Reading bKGD chunk length %d\n",len);
      if ((image->type==PALETTE)&&(len==1)){
	back=bbuff;
        *back=1;
	fread(back+1,1,len,in);
      }
      else if (((image->colour==GREY)||(image->colour==GREYA))&&(len==2)){
	back=bbuff;
        *back=1;
        fread(buff,1,len,in);
	if (channel_depth==16) back[1]=buff[0];
	else back[1]=buff[1];
      }
      else if (((image->colour==RGB)||(image->colour==RGBA))&&(len==6)){
	back=bbuff;
        back[0]=3;
        fread(buff,1,len,in);
	if (channel_depth==16){
	  back[1]=buff[0];
	  back[2]=buff[2];
	  back[3]=buff[4];
	}
	else{
	  back[1]=buff[1];
	  back[2]=buff[3];
	  back[3]=buff[5];
	}
      }
      else{
	fprintf(stderr,"Failed to understand bKGD chunk of %d bytes\n",len);
	for(i=0;i<len;i++){
	  if (fgetc(in)==EOF){
	    fprintf(stderr,"Unexpected EOF\n");
	    exit(1);
	  }
	}
      }
    }
    /* IDAT */
    else if ((buff[4]=='I')&&(buff[5]=='D')&&(buff[6]=='A')&&(buff[7]=='T')){
      iptr=realloc(iptr,ilen+len);
      if(!iptr){
	fprintf(stderr,"Realloc error for IDAT.\n");
	exit(1);
      }
      if (fread(iptr+ilen,1,len,in)!=len){
	fprintf(stderr,"Read error for %d bytes\n",len);
	exit(1);
      }
      ilen=ilen+len;
      if (debug>2)
	fprintf(stderr,"Read IDAT chunk of %d bytes, total %d bytes\n",
		len,ilen);
    }
    /* tRNS */
    else if ((buff[4]=='t')&&(buff[5]=='R')&&(buff[6]=='N')&&(buff[7]=='S')){
      {
	if (debug) fprintf(stderr,"Image contains transparency data\n");
        if (image->colour==GREY){
	  if (len!=2){ 
	    fprintf(stderr,"Unexpected length for tRNS chunk\n");
	    exit(1); 
	  } 
	  fread(buff+8,1,len,in);
	  image->trans=index;
	  image->trans[0]=buff[9];
	}
	else if (image->colour==RGB){
	  if (len!=6){
	    fprintf(stderr,"Unexpected length for tRNS chunk\n");
	    exit(1);
	  }
	  fread(buff+8,1,len,in);
	  image->trans=index;
	  image->trans[0]=buff[9];
	  image->trans[1]=buff[11];
	  image->trans[2]=buff[13];
	}
	else if (image->colour==PALETTE){
          image->apalette=malloc(len);
	  if(!image->apalette){
	    fprintf(stderr,"Malloc error for alpha palette\n");
	    exit(1);
	  }
	  image->apal_len=len;
	  fread(image->apalette,1,len,in);
	  if (debug)
	    fprintf(stderr,"Reading alpha palette of %d entries\n",len);

	  hit=-1;
	  for(i=0;i<len;i++){
	    if ((image->apalette[i]==0)&&(hit==-1)) hit=i;
            else if (image->apalette[i]!=255) hit=-2;
	  }
	  if (hit==-1){
	    if (debug)
	      fprintf(stderr,"No transparency found in alpha palette\n");
	    free(image->apalette);
	    image->apalette=NULL;
	    image->apal_len=0;
	  }
	  if (hit>=0){
	    if (debug)
	      fprintf(stderr,"Single palette entry has full transparency\n");
	    image->trans=index;
	    image->trans[0]=hit;
	    free(image->apalette);
	    image->apalette=NULL;
	    image->apal_len=0;
            if (debug) fprintf(stderr,"Colour %d is transparent\n",hit);
	  }
	}
	else{
	  fprintf(stderr,"Warning: impossible image->colour\n");
	  for(i=0;i<len;i++) fgetc(in);
	}
      }
    }
    /* tEXt */
    else if ((buff[4]=='t')&&(buff[5]=='E')&&(buff[6]=='X')&&(buff[7]=='t')){
      if (debug) fprintf(stderr,"Ignoring tEXt chunk of %d bytes",len);
      if (debug>1)
	fprintf(stderr," containing\n");
      else
	fprintf(stderr,"\n");
      for(i=0;i<len;i++){
	j=fgetc(in);
	if (j==EOF) error_exit("Unexpected EOF");
	if (debug>1) {
	  if (j) fputc(j,stderr);
	  else fprintf(stderr," : ");
	}
      }
      if (debug>1) fprintf(stderr,"\n");
    }
    /* IEND */
    else if ((buff[4]=='I')&&(buff[5]=='E')&&(buff[6]=='N')&&(buff[7]=='D'))
      break;
    else{
      if (debug>1) fprintf(stderr,"Ignoring %c%c%c%c chunk.\n",buff[4],buff[5],
			 buff[6],buff[7]);
      for(i=0;i<len;i++)
	if (fgetc(in)==EOF){
	  fprintf(stderr,"Unexpected EOF\n");
	  exit(1);
	}
    }
    /* Throw away CRC */
    fread(buff,4,1,in);
  }

  if(!ilen) {
    fprintf(stderr,"Error: no IDAT data found in PNG\n");
    exit(1);
  }

  if (debug>1) fprintf(stderr,"Read %d bytes of compressed data\n",ilen);

  if (back){
    if (image->bkgd==NULL) image->bkgd=back;
    else fprintf(stderr,"Ignoring bKGD from PNG file\n");
  }

  if ((!alpha)&&(image->flags&PNG_RAW)&&(!adam7)){
    image->data=iptr;
    image->data_len=ilen;
    image->compression=PNG;
    return;
  }

  if ((image->flags&PNG_RAW)&&(adam7)&&(image->flags&PNG)){
    image->data=iptr;
    image->data_len=ilen;
    image->compression=PNG_A7;
    return;
  }
  
#ifndef USE_ZLIB
  if ((alpha)&&((image->flags&PNG)==0)){
    fprintf(stderr,"Error: PNGs with alpha channel need a bmp2eps "
	    "compiled with zlib\n");
    exit(1);
  }
  if ((adam7)&&((image->flags&PNG)==0)){
    fprintf(stderr,"Error: interlaced PNGs need a bmp2eps "
	    "compiled with zlib\n");
    exit(1);
  }
  image->data=iptr;
  image->data_len=ilen;
  image->compression=PNG;

#else

  len=((image->width*image->depth+7)/8)*image->height;
  if (adam7){
    /* Add prediction bytes */
    len+=3*(image->height+7)/8+3*(image->height+3)/4+2*(image->height+1)/2;
    /* Unknown amount of extra padding (0 or 1 bytes)
       in the rows for six passes */
    if (image->depth<8)
      len+=3*(image->height+7)/8+3*(image->height+3)/4+(image->height+1)/2;
  }
  else{
    /* Add prediction bytes, one per line */
    len+=image->height;
  }
  
  dat=malloc(len);
  if (!dat){
    fprintf(stderr,"Malloc error for dat.\n");
    exit(1);
  }

  zlen=len;
  i=uncompress(dat,&zlen,iptr,ilen);
  if(i!=Z_OK){
    fprintf(stderr,"Decompression error, %d decompressing %d bytes to %d.\n",
                   i,(int)ilen,(int)len);
    exit(1);
  }
  free(iptr);

  bytes_per_pixel=(image->depth+7)/8;

  if (adam7)
    de_adam7(image,dat,channel_depth);
  else
    de_predict(image,dat,channel_depth);

  /* Reduce 16 bits per channel to eight */

  if (channel_depth==16){
    if (debug) fprintf(stderr,"Reducing from 16 bits per channel to 8\n");
    p=d=image->data;
    for(i=0;i<image->width*image->height*bytes_per_pixel/2;i++){
      *d++=*p++;
      p++;
    }
    channel_depth=8;
    bytes_per_pixel=bytes_per_pixel/2;
    image->depth=image->depth/2;
    image->data_len=(image->width*image->depth+7)/8*image->height;
    p=realloc(image->data,image->data_len);
    if (p) image->data=p;
  }
#endif
}

static void de_up(unsigned char *source, unsigned char *dest, 
		  unsigned char *last, unsigned int plen,
		  int bytes_per_pixel){
  int i,blen;

  blen=plen*bytes_per_pixel;

  for(i=0;i<blen;i++) *dest++=*(source++)+*last++;
}

static void de_sub(unsigned char *source, unsigned char *dest, 
		  unsigned int plen, int bytes_per_pixel){
  int i,blen;

  blen=plen*bytes_per_pixel;
  if (bytes_per_pixel==1){
    *dest++=*source++;
    for(i=1;i<plen;i++) {
      *dest=*(dest-1)+*source++;
      dest++;
    }
  }
  else{
    for(i=0;i<bytes_per_pixel;i++)
      *dest++=*source++;
    for(i=bytes_per_pixel;i<blen;i++) {
      *dest=*(dest-bytes_per_pixel)+*source++;
      dest++;
    }
  }
}

static void de_av(unsigned char *source, unsigned char *dest, 
		  unsigned char* last, unsigned int plen,
		  int bytes_per_pixel){
  int i,blen,av;

  blen=plen*bytes_per_pixel;

  if (last){
    if (bytes_per_pixel==1){
      *dest++=*(source)+*last++/2;
      for(i=1;i<plen;i++) {
	av=((int)(*(dest-1))+(int)(*last++))>>1;
	*dest++=*(source+1)+av;
	source++;
      }
    }
    else{
      for(i=0;i<bytes_per_pixel;i++)
        *dest++=*(source++)+*last++/2;
      for(i=bytes_per_pixel;i<blen;i++) {
	av=((int)(*(dest-bytes_per_pixel))+(int)(*last++))>>1;
	*dest++=*(source++)+av;
      }
    }
  }
  else{ /* Mad, but valid */
    if (bytes_per_pixel==1){
      *dest++=*(source);
      for(i=1;i<plen;i++) {
	av=((int)(*(dest-1)))>>1;
	*dest++=*(source+1)+av;
	source++;
      }
    }
    else{
      for(i=0;i<bytes_per_pixel;i++)
        *dest++=*source++;
      for(i=bytes_per_pixel;i<blen;i++) {
	av=((int)(*(dest-bytes_per_pixel)))>>1;
	*dest++=*(source++)+av;
      }
    }
  }
}

static void de_paeth(unsigned char *source, unsigned char *dest,
		     unsigned char *last,
		     unsigned int plen, int bytes_per_pixel){
  int i,blen,a,b,c,p,x;
  unsigned char *s2;

  blen=plen*bytes_per_pixel;
  s2=last;

  if (bytes_per_pixel==1){
    *dest++=*source+*s2++;
    for(i=1;i<plen;i++) {
      a=*(dest-1);
      b=*s2;
      c=*(s2-1);
      p=a+b-c;
      if ((abs(p-a)<=abs(p-b))&&(abs(p-a)<=abs(p-c))) x=a;
      else if (abs(p-b)<=abs(p-c)) x=b;
      else x=c;
      *dest++=*(source+1)+x;
      source++;
      s2++;
    }
  }
  else{
    for(i=0;i<bytes_per_pixel;i++)
      *dest++=*(source++)+*s2++;
    for(i=bytes_per_pixel;i<blen;i++) {
      a=*(dest-bytes_per_pixel);
      b=*s2;
      c=*(s2-bytes_per_pixel);
      p=a+b-c;
      if ((abs(p-a)<=abs(p-b))&&(abs(p-a)<=abs(p-c))) x=a;
      else if (abs(p-b)<=abs(p-c)) x=b;
      else x=c;
      *dest++=*(source++)+x;
      s2++;
    }
  }
}

/* Undo predictors */
static void de_predict(struct bitmap *image, unsigned char *dat,
		       int channel_depth)
{
  int debug,i,method,row_len;
  unsigned plen,bytes_per_pixel;
  int pred[5];
  unsigned char *p,*d,*l;

  debug=image->flags&DEBUG_MASK;

  p=dat;
  d=dat;
  l=NULL;
  pred[0]=pred[1]=pred[2]=pred[3]=pred[4]=0;
  row_len=(image->width*image->depth+7)/8;
  bytes_per_pixel=(image->depth+7)/8;
  plen=image->width;
  if (channel_depth<8) plen=(plen*channel_depth+7)/8;
  for(i=0;i<image->height;i++){
    if (debug>3) fprintf(stderr,"Row %5d predictor %d\n",i,*p);
    method=*p++;
    switch (method){
    case 0:
      memmove(d,p,row_len);
      break;
    case 1:
      de_sub(p,d,plen,bytes_per_pixel);
      break;
    case 2:
      if (l)
	de_up(p,d,l,plen,bytes_per_pixel);
      else
	memmove(d,p,row_len);
      break;
    case 3:
      de_av(p,d,l,plen,bytes_per_pixel);
      break;
    case 4:
      if (l)
	de_paeth(p,d,l,plen,bytes_per_pixel);
      else
	de_sub(p,d,plen,bytes_per_pixel);
      break;
    default:
      fprintf(stderr,"Invalid predictor in call to predict\n");
      exit(1);
    }
    l=d;
    p+=row_len;
    d+=row_len;
    pred[method]++;
  }

  image->data=dat;
  image->data_len=(image->width*image->depth+7)/8*image->height;

  if (debug>1) {
    fprintf(stderr,"Predictors used: ");
    for(i=0;i<5;i++) fprintf(stderr,"%dx%d ",i,pred[i]);
    fprintf(stderr,"\n");
  }

}

/* image - filled with result, width, height, depth already set
 *  dat - uncompressed data
 *  channel_depth - bits per channel
 *
 * For each Adam7 "pass",
 *   fill in the structure pass to reflect that subimage
 *   call de_predict with that pass
 *   copy data out to final image data
 *   advance pointer to input data, p, appropriately
 * 
 */

static void de_adam7(struct bitmap *image, unsigned char *dat,
		       int channel_depth){
  int debug;
  unsigned int i,j,k,row_len,pass_len,bytes_per_pixel;
  struct bitmap pass;
  unsigned char *s,*d,*p;

  debug=image->flags&DEBUG_MASK;

  if (debug) fprintf(stderr,
                     "Deinterlacing PNG with multiple pixels per byte\n");

  if (image->depth<8){
    de_adam7_bits(image,dat,channel_depth);
    return;
  }

  row_len=(image->width*image->depth+7)/8;
  image->data_len=row_len*image->height;
  image->data=malloc(image->data_len);
  if(!image->data){
    fprintf(stderr,"Malloc error in de_adam7\n");
    exit(1);
  }
#if 0
  p=image->data;
  for(i=0;i<image->data_len;i++) *(p++)=0;
#endif

  pass.depth=image->depth;
  pass.flags=image->flags;
  bytes_per_pixel=image->depth/8;
  p=dat;


  /* pass one */

  pass.width=(image->width+7)/8;
  pass_len=pass.width*bytes_per_pixel;
  pass.height=(image->height+7)/8;
  de_predict(&pass,p,channel_depth);
  for(i=0;i<pass.height;i++){
    for(j=0;j<pass.width;j++){
      s=p+i*pass_len+j*image->depth/8;
      d=image->data+8*i*row_len+8*j*bytes_per_pixel;
      for(k=0;k<bytes_per_pixel;k++)
	*(d++)=*(s++);
    }
  }
  p+=(pass_len+1)*pass.height;

  /* pass two */

  pass.width=(image->width+3)/8;
  pass_len=pass.width*bytes_per_pixel;
  pass.height=(image->height+7)/8;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*bytes_per_pixel;
	d=image->data+8*i*row_len+(8*j+4)*bytes_per_pixel;
	for(k=0;k<bytes_per_pixel;k++)
	  *(d++)=*(s++);
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  /* pass three */

  pass.width=(image->width+3)/4;
  pass_len=pass.width*bytes_per_pixel;
  pass.height=(image->height+3)/8;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*bytes_per_pixel;
	d=image->data+(8*i+4)*row_len+4*j*bytes_per_pixel;
	for(k=0;k<bytes_per_pixel;k++)
	  *(d++)=*(s++);
      }
    }
    p+=(pass_len+1)*pass.height;
  }


  /* pass four */

  pass.width=(image->width+1)/4;
  pass_len=pass.width*bytes_per_pixel;
  pass.height=(image->height+3)/4;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*bytes_per_pixel;
	d=image->data+4*i*row_len+(4*j+2)*bytes_per_pixel;
	for(k=0;k<bytes_per_pixel;k++)
	  *(d++)=*(s++);
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  /* pass five */

  pass.width=(image->width+1)/2;
  pass_len=pass.width*bytes_per_pixel;
  pass.height=(image->height+1)/4;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*bytes_per_pixel;
	d=image->data+(4*i+2)*row_len+2*j*bytes_per_pixel;
	for(k=0;k<bytes_per_pixel;k++)
	  *(d++)=*(s++);
      }
    }
    p+=(pass_len+1)*pass.height;
  }


  /* pass six */

  pass.width=image->width/2;
  pass_len=pass.width*bytes_per_pixel;
  pass.height=(image->height+1)/2;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*bytes_per_pixel;
	d=image->data+2*i*row_len+(2*j+1)*bytes_per_pixel;
	for(k=0;k<bytes_per_pixel;k++)
	  *(d++)=*(s++);
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  /* pass seven */

  pass.width=image->width;
  pass_len=pass.width*bytes_per_pixel;
  pass.height=image->height/2;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*bytes_per_pixel;
	d=image->data+(2*i+1)*row_len+j*bytes_per_pixel;
	for(k=0;k<bytes_per_pixel;k++)
	  *(d++)=*(s++);
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  free(dat);

}

/* image - filled with result, width, height, depth already set
 *  dat - uncompressed data
 *  channel_depth - bits per channel
 *
 * routine for bits per pixel of 1, 2, 4 and 8.
 * Produces temporary output at one byte per pixel, then
 * call pre-existing compact routine.
 */

static void de_adam7_bits(struct bitmap *image, unsigned char *dat,
			  int channel_depth){
  int debug;
  unsigned int i,j,pass_len,mask,pixels_per_byte;
  unsigned int bits_per_pixel;
  struct bitmap pass;
  unsigned char *s,*d,*p,c;

  debug=image->flags&DEBUG_MASK;

  if (debug) fprintf(stderr,"Deinterlacing PNG\n");

  if ((image->depth!=4)&&(image->depth!=2)&&(image->depth!=1)){
    fprintf(stderr,
	    "Unable to deinterlace image depth of %d\n",image->depth);
    exit(1);
  }

  image->data_len=image->width*image->height;
  image->data=malloc(image->data_len);
  if(!image->data){
    fprintf(stderr,"Malloc error in de_adam7_bits\n");
    exit(1);
  }
#if 0
  p=image->data;
  for(i=0;i<image->data_len;i++) *(p++)=0;
#endif

  pass.depth=image->depth;
  pass.flags=image->flags;
  bits_per_pixel=image->depth;
  pixels_per_byte=8/image->depth;
  mask=(1<<image->depth)-1;
  p=dat;


  /* pass one */

  pass.width=(image->width+7)/8;
  pass_len=(pass.width*image->depth+7)/8;
  pass.height=(image->height+7)/8;
  de_predict(&pass,p,channel_depth);
  for(i=0;i<pass.height;i++){
    for(j=0;j<pass.width;j++){
      s=p+i*pass_len+j*image->depth/8;
      c=((*s)>>(8-bits_per_pixel*(j%pixels_per_byte+1)))&mask;
      d=image->data+8*i*image->width+8*j;
      *d=c;
    }
  }
  p+=(pass_len+1)*pass.height;

  /* pass two */

  pass.width=(image->width+3)/8;
  pass_len=(pass.width*image->depth+7)/8;
  pass.height=(image->height+7)/8;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*image->depth/8;
	c=((*s)>>(8-bits_per_pixel*(j%pixels_per_byte+1)))&mask;
	d=image->data+8*i*image->width+(8*j+4);
	*d=c;
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  /* pass three */

  pass.width=(image->width+3)/4;
  pass_len=(pass.width*image->depth+7)/8;
  pass.height=(image->height+3)/8;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*image->depth/8;
	c=((*s)>>(8-bits_per_pixel*(j%pixels_per_byte+1)))&mask;
	d=image->data+(8*i+4)*image->width+4*j;
	*d=c;
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  /* pass four */

  pass.width=(image->width+1)/4;
  pass_len=(pass.width*image->depth+7)/8;
  pass.height=(image->height+3)/4;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*image->depth/8;
	c=((*s)>>(8-bits_per_pixel*(j%pixels_per_byte+1)))&mask;
	d=image->data+4*i*image->width+(4*j+2);
        *d=c;
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  /* pass five */

  pass.width=(image->width+1)/2;
  pass_len=(pass.width*image->depth+7)/8;
  pass.height=(image->height+1)/4;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*image->depth/8;
	c=((*s)>>(8-bits_per_pixel*(j%pixels_per_byte+1)))&mask;
	d=image->data+(4*i+2)*image->width+2*j;
        *d=c;
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  /* pass six */

  pass.width=image->width/2;
  pass_len=(pass.width*image->depth+7)/8;
  pass.height=(image->height+1)/2;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*image->depth/8;
	c=((*s)>>(8-bits_per_pixel*(j%pixels_per_byte+1)))&mask;
	d=image->data+2*i*image->width+(2*j+1);
	*d=c;
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  /* pass seven */

  pass.width=image->width;
  pass_len=(pass.width*image->depth+7)/8;
  pass.height=image->height/2;
  if (pass.width&&pass.height){
    de_predict(&pass,p,channel_depth);
    for(i=0;i<pass.height;i++){
      for(j=0;j<pass.width;j++){
	s=p+i*pass_len+j*image->depth/8;
	c=((*s)>>(8-bits_per_pixel*(j%pixels_per_byte+1)))&mask;
	d=image->data+(2*i+1)*image->width+j;
	*d=c;
      }
    }
    p+=(pass_len+1)*pass.height;
  }

  image->pal_len=1<<image->depth;
  compact(image);

  free(dat);

}

