/* Copyright (c) 2010, 2019 MJ Rutter 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street,
 * Fifth Floor, Boston, MA  02110-1301, USA.
 */ 

/* Read a TGA file
 *
 * As TGA haas no magic header, we need to be passed the first
 * three bytes which are usually swallowed as containing pure magic
 *
 * Assert that 16 bit TGA files are one bit alpha, and five each of b, g and r
 * NOT 6 of g as with some other formats
 *
 * The extra fields of TGA v2.0 are ignored too: they contain little of
 * interest, and are nearly impossible to find without using seek(), and
 * we must be able to read from stdin.
 */

#include <stdio.h>
#include <stdlib.h>  /* malloc */
#include "bmp2eps.h"

static void tga_row_read(FILE* file, int width, int depth, int rle,
			 unsigned char* data);

static void bgr2rgb(int size, unsigned char *data);
static void bgra2rgba(int size, unsigned char *data);

void tgaread(FILE *in, struct bitmap *image, unsigned char* buff){
  unsigned char header[18],junk[255],*ptr,*ptr2;
  unsigned width, height, pal_off, pal_len, pal_depth;
  int debug,depth,i,bytes_per_pixel,rle=0,invert,alpha;

  debug=image->flags&DEBUG_MASK;

  if (debug) fprintf(stderr,"TGA file suspected\n");

  header[0]=buff[0];
  header[1]=buff[1];
  header[2]=buff[2];

/* Read Header */
 
  if (!fread(header+3,15,1,in)){
    fprintf(stderr,"Error reading TGA header\n");
    exit(1);
  }
  

  width=header[12]+256*header[13];
  height=header[14]+256*header[15];
  depth=header[16];

  if (debug>1) fprintf(stderr,"TGA width %d, height %d, depth %d\n",
                       width, height,depth);

  if (header[17]&0xc0){
    fprintf(stderr,"Error in TGA header: bits 6 & 7 of last byte non zero.\n");
    if (debug) fprintf(stderr,"Errant byte is %x\n",(int)header[17]);
    exit(1);
  }

  alpha=header[17]&0xf;
  if ((alpha==8)&&(depth!=32))
    error_exit("TGA with alpha=8 but depth!=32");
  else if ((alpha==1)&&((depth!=15)&&(depth!=16)))
    error_exit("TGA with alpha=1 but depth!=15 or 16");

  if ((alpha==0)&&(depth==16)) depth=15;
  if ((alpha==1)&&(depth==15)) depth=16;
  if ((debug>1)&&((depth==15)||(depth==16)))
    fprintf(stderr,"Alpha bit%s present\n",alpha?"":" not");
  
/* Skip optional ID field, whose length is given by the first byte */

  if (debug>1) fprintf(stderr,"ID field has length %d\n",(int)header[0]);

  fread(junk,header[0],1,in);

  if(header[1]==1){ /* We have a pallete */
    pal_off=header[3]+256*header[4];
    pal_len=header[5]+256*header[6];
    if (pal_off+pal_len>256){
      fprintf(stderr,"Palletted TGA of %d palette entries not supported\n",
         pal_off+pal_len);
      exit(1);
    }
    if(!(image->palette=malloc(4*(pal_off+pal_len)))){
      fprintf(stderr,"Malloc error for palette in tgaread\n");
      exit(1);
    }
    pal_depth=header[7];
    if ((pal_depth!=15)&&
        (pal_depth!=16)&&
        (pal_depth!=24)&&
        (pal_depth!=32)){
      fprintf(stderr,"Error, palette depth of %d not supported\n",pal_depth);
      exit(1);
    }

    if (debug>1) fprintf(stderr,"TGA is paletted, palette depth %d, "
                         "entries %d, used entries %d\n",pal_depth,
                         pal_len+pal_off,pal_len);
/* Palette data are stored identically to an uncompressed row */
    if ((pal_depth!=32)&&(pal_depth!=16)){
      for(i=0;i<3*pal_off;i++)
	image->palette[i]=0;
      tga_row_read(in,pal_len,pal_depth,0,image->palette+3*pal_off);
    }
    else{
      for(i=0;i<4*pal_off;i++)
	image->palette[i]=0;
      tga_row_read(in,pal_len,pal_depth,0,image->palette+4*pal_off);
    }
    image->pal_len=pal_len+pal_off;

    if ((pal_depth==32)||(pal_depth==16)){
      if(!(image->apalette=malloc(pal_off+pal_len))){
	fprintf(stderr,"Malloc error for alpha palette in tgaread\n");
	exit(1);
      }
      ptr=image->palette;
      ptr2=image->palette;
      for(i=0;i<pal_off+pal_len;i++){
	*(ptr++)=*(ptr2++);
	*(ptr++)=*(ptr2++);
	*(ptr++)=*(ptr2++);
	image->apalette[i]=*(ptr2++);
      }
      image->apal_len=pal_off+pal_len;
    }

  }

  if (header[17]&0x10){
    fprintf(stderr,"Right to left TGAs not yet supported\n");
    exit(1);
  }

  switch(depth){
  case 8:
    bytes_per_pixel=1;
    break;
  case 15: case 24:
    bytes_per_pixel=3;
    break;
  case 16: case 32:
    bytes_per_pixel=4;
    break;
  default:
    fprintf(stderr,"TGA of depth %d not yet supported\n",depth);
    exit(1);
  }

  if(!(image->data=malloc(width*height*bytes_per_pixel))){
    fprintf(stderr,"Malloc error on raw in tgaread\n");
    exit(1);
  }

  image->height=height;
  image->width=width;
  image->data_len=width*height*bytes_per_pixel;
  image->depth=8*bytes_per_pixel;
  rle=(header[2]&8)>>3;
  invert=1-((header[17]&0x20)>>5);

  if (debug) fprintf(stderr,"TGA is%s run length encoded\n",rle?"":" not");
  if (debug) fprintf(stderr,"TGA is%s inverted\n",invert?"":" not");

  if (invert) ptr=image->data+width*(height-1)*bytes_per_pixel;
  else ptr=image->data;
  for(i=0;i<height;i++){
    tga_row_read(in,width,depth,rle,ptr);
    if (invert) ptr-=width*bytes_per_pixel;
    else ptr+=width*bytes_per_pixel;
  }
  switch(header[2]&3){
  case 1:
    image->type=PALETTE;
    image->colour=PALETTE;
    break;
  case 2:
    image->type=BMP;
    image->colour=RGB;
    if (bytes_per_pixel==4) image->colour=RGBA;
    break;
  case 3:
    image->type=BMP;
    image->colour=GREY;
    break;
  }

  return;
}

static void tga_row_read(FILE* file, int width, int depth, int rle,
			 unsigned char* data){
  unsigned char *ptr, *ptr0, *end, code, p[4];
  unsigned bytes_per_pixel,i,j,len,tmp;

  if ((rle==0)&&((depth==8)||(depth==24)||(depth==32))){
      if (!fread(data,width*depth/8,1,file)){
        fprintf(stderr,"Unexpected EOF reading TGA\n");
        exit(1);
      }
      if (depth==24) bgr2rgb(width,data);
      if (depth==32) bgra2rgba(width,data);
      return;
  }

  bytes_per_pixel=(depth+1)/8;
  if (bytes_per_pixel>4){
    fprintf(stderr,"Unsupported depth in tga_row_read\n");
    exit(1);
  }

  ptr0=ptr=data;

#if 0
  fprintf(stderr,"tgarr: width=%d, depth=%d, rle=%d, data=%p, ptr=%p, "
                 "bytes_per_pixel=%d\n",width,depth,rle,data,ptr,
                 bytes_per_pixel);
#endif

  if (rle){
    end=ptr+width*bytes_per_pixel;
    while(ptr<end){
      fread(&code,1,1,file);
      if ((code&0x80)==0){ /* Literal pixels */
        len=(code+1)*bytes_per_pixel;
        if (ptr+len>end){
          fprintf(stderr,"Error decoding RLE literal data\n");
          exit(1);
        }
        if (!fread(ptr,len,1,file)){
          fprintf(stderr,"Unexpected EOF reading TGA\n");
          exit(1);
        }
        ptr+=len;
      }else{ /* A run */
        if (!fread(p,bytes_per_pixel,1,file)){
          fprintf(stderr,"Unexpected EOF reading TGA\n");
          exit(1);
        }
        len=(code&0x7f)+1;
        if (ptr+len*bytes_per_pixel>end){
          fprintf(stderr,"Error decoding RLE run\n");
          exit(1);
        }
        for(i=0;i<len;i++)
          for(j=0;j<bytes_per_pixel;j++)
            *(ptr++)=p[j];
      }
    }
    if (ptr!=end){
      fprintf(stderr,"Error decoding RLE row\n");
      exit(1);
    }
  }else{
    if (!fread(ptr,width*((depth+1)/8),1,file)){
      fprintf(stderr,"Unexpected EOF reading TGA\n");
      exit(1);
    }
  }

  ptr=ptr0;

  switch(depth){
  case 8:
    break;
  case 24:
    bgr2rgb(width,data);
    break;
  case 32:
    bgra2rgba(width,data);
    break;
  case 15:
  case 16: /* Need to do this backwards to avoid overwriting input */
    ptr+=2*(width-1);
    data+=3*width-1;
    if (depth==16) data+=width;
    for(i=0;i<width;i++){
      tmp=*ptr+256*ptr[1];
      ptr-=2;
      if (depth==16){
	if (tmp&0x8000)
	  *(data--)=255;
	else
	  *(data--)=0;
      }
      *(data--)=(tmp&0x1f)<<3;
      *(data--)=(tmp&0x3e0)>>2;
      *(data--)=(tmp&0x7c00)>>7;
    }
  }

}

static void bgr2rgb(int size, unsigned char *data){
  int i;
  unsigned char *ptr,tmp;

  ptr=data;
  for(i=0;i<size;i++){
    tmp=*ptr;
    *ptr=*(ptr+2);
    *(ptr+2)=tmp;
    ptr+=3;
  }
}

static void bgra2rgba(int size, unsigned char *data){
  int i;
  unsigned char *ptr,tmp;

  ptr=data;
  for(i=0;i<size;i++){
    tmp=*ptr;
    *ptr=*(ptr+2);
    *(ptr+2)=tmp;
    ptr+=4;
  }
}
